// GM_MapWindow.cpp - interface for GM_MapWindow class
//

#include "StdAfx.h"
#include "GM_MapWindow.h"

#include "GlobalMapperInterface_3D.h"

#include <algorithm>
#include <tchar.h>

/*--------------------------------------------------------------------
                         STATIC FUNCTIONS
--------------------------------------------------------------------*/

//===================================================================
//
// FUNCTION:
//   Change3DViewCallback
//
// DESCRIPTION:
//   Callback function for when the 3D view window requests a new
//   view to be displayed.
//
//===================================================================
static void _stdcall 
Change3DViewCallback
    ( 
    const GM_Rectangle_t*   aNewViewBounds,
    void*                   aUserData           // user data passed into GM_3DSetChangeViewCallback function
    )
{
    // The user data should be a GM_MapWindow pointer
    GM_MapWindow* theMapWindow = (GM_MapWindow*)aUserData;

    // Update the view rectangle
    if ( ( NULL != theMapWindow ) && ( NULL != aNewViewBounds ) )
    {
        theMapWindow->SetCurrentViewRect( *aNewViewBounds );
    }
} // end of function 3DChangeViewCallbackFunc

/*--------------------------------------------------------------------
                          PUBLIC METHODS
--------------------------------------------------------------------*/

//===================================================================
//
// METHOD:
//   GM_MapWindow::GM_MapWindow
//
// DESCRIPTION:
//   Constructor for GM_MapWindow class.
//
//===================================================================
GM_MapWindow::GM_MapWindow()
: mDrawBgBitmap( NULL ),
  mDrawWnd( NULL ),
  mGridDisplay( eNoGrid ),
  mOrientation( 0.0 )
{
    memset( &mCurViewRect, 0, sizeof mCurViewRect );
    memset( &mDrawWndRect, 0, sizeof mDrawWndRect );
    memset( &mLast3DViewRect, 0, sizeof mLast3DViewRect );
} // end of function GM_MapWindow::GM_MapWindow

//===================================================================
//
// METHOD:
//   GM_MapWindow::~GM_MapWindow
//
// DESCRIPTION:
//   Destructor for GM_MapWindow class.
//
//===================================================================
GM_MapWindow::~GM_MapWindow()
{
    UnloadAllLayers( false );

    // Close the 3D window if one is open
    if ( GM_3DIsWindowOpen() )
    {
        GM_3DCloseViewWindow();
    }

    // Destroy the offscreen bitmap
    if ( NULL != mDrawBgBitmap )
    {
        ::DeleteObject( mDrawBgBitmap );
        mDrawBgBitmap = NULL;
    }
} // end of function GM_MapWindow::GM_MapWindow

//===================================================================
//
// METHOD:
//   GM_MapWindow::CenterOnGroundPoint
//
// DESCRIPTION:
//   Centers the map on the given ground point.
//
//===================================================================
void
GM_MapWindow::CenterOnGroundPoint
    (
    double  aX,
    double  aY
    )
{
    // Get the width and height of the view rect
    double theWidth = mCurViewRect.mMaxX - mCurViewRect.mMinX;
    double theHeight = mCurViewRect.mMaxY - mCurViewRect.mMinY;

    // Update the view rectangle to be centered on the new point
    mCurViewRect.mMinX = aX - theWidth / 2.0;
    mCurViewRect.mMaxX = mCurViewRect.mMinX + theWidth;
    mCurViewRect.mMinY = aY - theHeight / 2.0;
    mCurViewRect.mMaxY = mCurViewRect.mMinY + theHeight;

    // Redraw
    Invalidate();
} // end of function GM_MapWindow::CenterOnGroundPoint

//===================================================================
//
// METHOD:
//   GM_MapWindow::CenterOnPixel
//
// DESCRIPTION:
//   Centers the map on the given pixel location. Returns 
//   GM_Error_None if the view was successfully recentered.
//
//===================================================================
GM_Error_t32
GM_MapWindow::CenterOnPixel
    (
    sint32  aPixelX,
    sint32  aPixelY
    )
{
    // Convert the pixel coordinate to ground coordinates
    double theGroundX, theGroundY;
    GM_Error_t32 theErr = ConvertCoordPixelToGround
        (
        aPixelX,
        aPixelY,
        theGroundX,
        theGroundY
        );

    // Center on the new ground point
    if ( GM_Error_None == theErr )
    {
        CenterOnGroundPoint( theGroundX, theGroundY );
    }

    return ( theErr );
} // end of function GM_MapWindow::CenterOnPixel

//===================================================================
//
// METHOD:
//   GM_MapWindow::ConvertCoordGroundToPixel
//
// DESCRIPTION:
//   Converts a coordinate from the current ground coordinate system
//   to pixel coordinates referenced to the map draw window. Returns
//   GM_Error_None if the conversion is successful.
//
//===================================================================
GM_Error_t32
GM_MapWindow::ConvertCoordGroundToPixel
    (
    double  aGroundX,   // in: ground X coordinate
    double  aGroundY,   // in: ground Y coordinate
    sint32& aPixelX,    // out: x pixel coordinate in draw window
    sint32& aPixelY     // out: y pixel coordiante in draw window
    ) const
{
    // Convert the draw rectangle into a GM_PixelRect_t
    GM_PixelRect_t theDrawRect;
    theDrawRect.mLeft = mDrawWndRect.left;
    theDrawRect.mRight = mDrawWndRect.right;
    theDrawRect.mTop = mDrawWndRect.top;
    theDrawRect.mBottom = mDrawWndRect.bottom;

    // Set the orientation
    GM_SetOrientation( mOrientation );

    // Convert the coordinate to pixel coordinates
    GM_Error_t32 theErr = GM_ConvertCoordGroundToPixel
        (
        aGroundX,
        aGroundY,
        &aPixelX,
        &aPixelY,
        &mCurViewRect,
        &theDrawRect
        );

    return ( theErr );
} // end of function GM_MapWindow::ConvertCoordGroundToPixel

//===================================================================
//
// METHOD:
//   GM_MapWindow::ConvertCoordLatLonToPixel
//
// DESCRIPTION:
//   Converts a latitude/longitude coordinate with the current datum
//   to pixel coordinates referenced to the map draw window. Returns
//   GM_Error_None if the conversion is successful.
//
//===================================================================
GM_Error_t32
GM_MapWindow::ConvertCoordLatLonToPixel
    (
    double  aLat,       // in: latitude in degrees in current view datum
    double  aLon,       // in: longitude in degrees in current view datum
    sint32& aPixelX,    // out: x pixel coordinate in draw window
    sint32& aPixelY     // out: y pixel coordiante in draw window
    ) const
{
    // Get the current view projection
    GM_Projection_t theCurProj;
    GM_Error_t32 theErr = GM_GetProjection( &theCurProj );
    if ( GM_Error_None != theErr )
    {
        return ( theErr );
    }

    // Setup a lat/lon projection with the current datum
    GM_Projection_t theGeoProj;
    memset( &theGeoProj, 0, sizeof theGeoProj );
    theGeoProj.mProjSys = GM_PRJ_GEO;
    theGeoProj.mDatum = theCurProj.mDatum;
    theGeoProj.mUnit = GM_PRJ_UNIT_ARC_DEGREES;

    // Project the lat/lon to ground coordinates
    double theGroundX, theGroundY;
    theErr = GM_ProjectPoint
        ( 
        aLon, aLat, &theGroundX, &theGroundY, &theGeoProj, NULL
        );
    if ( GM_Error_None != theErr )
    {
        return ( theErr );
    }

    // Convert the ground coordinates to pixel
    return ( ConvertCoordGroundToPixel( theGroundX, theGroundY, aPixelX, aPixelY ) );
} // end of function GM_MapWindow::ConvertCoordLatLonToPixel

//===================================================================
//
// METHOD:
//   GM_MapWindow::ConvertCoordPixelToGround
//
// DESCRIPTION:
//   Converts a pixel coordinate to the current ground coordinate 
//   system. Returns GM_Error_None if the conversion is successful.
//
//===================================================================
GM_Error_t32
GM_MapWindow::ConvertCoordPixelToGround
    (
    sint32  aPixelX,    // in: x pixel coordinate in draw window
    sint32  aPixelY,    // in: y pixel coordiante in draw window
    double& aGroundX,   // out: ground X coordinate
    double& aGroundY    // out: ground Y coordinate
    ) const
{
    // Convert the draw rectangle into a GM_PixelRect_t
    GM_PixelRect_t theDrawRect;
    theDrawRect.mLeft = mDrawWndRect.left;
    theDrawRect.mRight = mDrawWndRect.right;
    theDrawRect.mTop = mDrawWndRect.top;
    theDrawRect.mBottom = mDrawWndRect.bottom;

    // Set the orientation
    GM_SetOrientation( mOrientation );

    // Convert the coordinate to pixel coordinates
    GM_Error_t32 theErr = GM_ConvertCoordPixelToGround
        (
        aPixelX,
        aPixelY,
        &aGroundX,
        &aGroundY,
        &mCurViewRect,
        &theDrawRect
        );

    return ( theErr );
} // end of function GM_MapWindow::ConvertCoordPixelToGround

//===================================================================
//
// METHOD:
//   GM_MapWindow::ConvertCoordPixelToLatLon
//
// DESCRIPTION:
//   Converts a pixel coordinate to a latitude/longitude coordinate 
//   with the current datum. Returns GM_Error_None if the conversion 
//   is successful.
//
//===================================================================
GM_Error_t32
GM_MapWindow::ConvertCoordPixelToLatLon
    (
    sint32  aPixelX,    // in: x pixel coordinate in draw window
    sint32  aPixelY,    // in: y pixel coordiante in draw window
    double& aLat,       // out: latitude in degrees in current view datum
    double& aLon        // out: longitude in degrees in current view datum
    ) const
{
    // Get the current view projection
    GM_Projection_t theCurProj;
    GM_Error_t32 theErr = GM_GetProjection( &theCurProj );
    if ( GM_Error_None != theErr )
    {
        return ( theErr );
    }

    // Setup a lat/lon projection with the current datum
    GM_Projection_t theGeoProj;
    memset( &theGeoProj, 0, sizeof theGeoProj );
    theGeoProj.mProjSys = GM_PRJ_GEO;
    theGeoProj.mDatum = theCurProj.mDatum;
    theGeoProj.mUnit = GM_PRJ_UNIT_ARC_DEGREES;

    // Get the ground coordinates associated with the pixel
    double theGroundX, theGroundY;
    theErr = ConvertCoordPixelToGround( aPixelX, aPixelY, theGroundX, theGroundY );
    if ( GM_Error_None != theErr )
    {
        return ( theErr );
    }

    // Project the ground coordinates to lat/lon
    theErr = GM_ProjectPoint
        ( 
        theGroundX, theGroundY, &aLon, &aLat, NULL, &theGeoProj
        );
    return ( theErr );
} // end of function GM_MapWindow::ConvertCoordPixelToLatLon

//===================================================================
//
// METHOD:
//   GM_MapWindow::Draw
//
// DESCRIPTION:
//   Draws the loaded layers to the given device context using
//   the provided draw order. You must have called SetDrawWindowRect
//   at some point prior to calling this function.
//
//===================================================================
GM_Error_t32
GM_MapWindow::Draw
    (
    HDC                 aDC,
    COLORREF            aBackgroundColor,
    DrawOrder_t         aDrawOrder,
    bool                aEraseBackground,
    GM_DrawFlags_t32    aDrawFlags
    )
{
    GM_Error_t32 theErr = GM_Error_None;

    // Make sure a real device context was passed in
    if ( NULL == aDC )
    {
        return ( GM_Error_InvalidParam );
    }

    // Create a memory device context to draw to
    HDC theMemDC = ::CreateCompatibleDC( aDC );
    if ( NULL == theMemDC )
    {
        ShowMessageBox( _T("Error creating memory device context to draw to.") );
        return ( GM_Error_DrawError );
    }

    // If we need to do a full draw, allocate the offscreen bitmap and draw
    bool bDrawNeeded = ( NULL == mDrawBgBitmap );
    if ( NULL == mDrawBgBitmap )
    {
        // Allocate the offscreen bitmap
        mDrawBgBitmap = ::CreateCompatibleBitmap
            (
            aDC, 
            mDrawWndRect.right - mDrawWndRect.left,
            mDrawWndRect.bottom - mDrawWndRect.top
            );
    }

    // If there is a valid background bitmap, just blit from that.
    if ( NULL != mDrawBgBitmap )
    {
        // Select the background bitmap into the memory DC
        HGDIOBJ theOldBitmap = ::SelectObject( theMemDC, mDrawBgBitmap );

        // Draw if necessary
        if ( bDrawNeeded )
        {
            theErr = DrawToOffscreenBuffer
                (
                theMemDC,
                aBackgroundColor,
                aDrawOrder,
                aEraseBackground,
                aDrawFlags
                );
        }

        // BitBlt from the memory DC to the screen
        ::BitBlt
            (
            aDC,
            mDrawWndRect.left,
            mDrawWndRect.top,
            mDrawWndRect.right - mDrawWndRect.left,
            mDrawWndRect.bottom - mDrawWndRect.top,
            theMemDC,
            0,
            0,
            SRCCOPY
            );

        // Select the old bitmap back
        ::SelectObject( theMemDC, theOldBitmap );
    }
    else
    {
        theErr = GM_Error_DrawError;
    }

    // Free the memory DC
    ::DeleteObject( theMemDC );

    return ( theErr );
} // end of function GM_MapWindow::Draw

//===================================================================
//
// METHOD:
//   GM_MapWindow::FindNearestFeatures
//
// DESCRIPTION:
//   Finds the nearest loaded (and drawn) vector feature(s) to the
//   given pixel location. A list of features matching the search
//   criteria will be returned.
//
//===================================================================
GM_MapWindow::FoundFeatureList_t
GM_MapWindow::FindNearestFeatures
    (
    const POINT&        aSearchPt,              // pixel location to search at
    uint32              aSearchRadius,          // search radius in screen pixels
    uint32              aMaxFeatures,           // max number of features to find
    GM_FindFlags_t32    aFindFlags
    )
{
    FoundFeatureList_t theFoundFeatureList;

    // Allocate a search buffer for the results
    GM_FoundFeature_t* theResultsList = new GM_FoundFeature_t[ aMaxFeatures ];
    if ( NULL == theResultsList )
    {
        return ( theFoundFeatureList );
    }

    // Convert the draw rectangle into a GM_PixelRect_t
    GM_PixelRect_t theDrawRect;
    theDrawRect.mLeft = mDrawWndRect.left;
    theDrawRect.mRight = mDrawWndRect.right;
    theDrawRect.mTop = mDrawWndRect.top;
    theDrawRect.mBottom = mDrawWndRect.bottom;

    // Do the search
    uint32 theFoundCnt = 0;
    GM_Point_t theSearchPt;
    theSearchPt.mX = aSearchPt.x;
    theSearchPt.mY = aSearchPt.y;
    GM_Error_t32 theRetVal = GM_FindNearestFeatures
        (
        &theSearchPt,
        NULL,                       // search all loaded vector layers
        0,  
        aFindFlags,
        &mCurViewRect,
        &theDrawRect,
        theResultsList,
        aMaxFeatures,
        &theFoundCnt,
        aSearchRadius
        );

    // Add the results to our list
    for ( uint32 i = 0; i < theFoundCnt; i++ )
    {
        theFoundFeatureList.push_back( theResultsList[i] );
    }

    // Free the results list buffer
    delete[] theResultsList;
    theResultsList = NULL;

    return ( theFoundFeatureList );
} // end of function GM_MapWindow::FindNearestFeatures

//===================================================================
//
// METHOD:
//   GM_MapWindow::GetRasterDisplayOptions
//
// DESCRIPTION:
//   Gets the current raster display options.
//
//===================================================================
GM_Error_t32
GM_MapWindow::GetRasterDisplayOptions
    (
    GM_LayerHandle_t32          aLayer,
    GM_RasterDisplayOptions_t&  aDisplayOpts
    ) const
{
    memset( &aDisplayOpts, 0, sizeof aDisplayOpts );
    aDisplayOpts.mSize = sizeof( GM_RasterDisplayOptions_t );
    return ( GM_GetRasterDisplayOptions( aLayer, &aDisplayOpts ) );
} // end of function GM_MapWindow::GetRasterDisplayOptions

//===================================================================
//
// METHOD:
//   GM_MapWindow::Hide3DView
//
// DESCRIPTION:
//   Hides the 3D view window, if any.
//
//===================================================================
void
GM_MapWindow::Hide3DView()
{
    GM_3DCloseViewWindow();
    memset( &mLast3DViewRect, 0, sizeof mLast3DViewRect );
} // end of function GM_MapWindow::Hide3DView

//===================================================================
//
// METHOD:
//   GM_MapWindow::LoadGenericAsciiTextLayer
//
// DESCRIPTION:
//   Loads a layer from the given ASCII text file with the specified
//   format.
//
//   If <aLoadedLayer> is not NULL, the handle for the loaded layer
//   is saved.
//
//===================================================================
GM_Error_t32
GM_MapWindow::LoadGenericAsciiTextLayer
    (
    LPCTSTR                     aFilename,
    const GM_AsciiFormatInfo_t* aFormatInfo,
    GM_LoadFlags_t32            aLoadFlags,
    GM_LayerHandle_t32*         aLoadedLayer,
    bool                        aZoomToAllLayers
    )
{
    // If we're in a unicode build, get a multi-byte version of the filename
    #ifdef _UNICODE
    char* theFilename = ConvertUnicodeToMultibyte( aFilename );
    if ( NULL == theFilename )
    {
        return ( GM_Error_OutOfMemory );
    }
    #else
    const char* theFilename = aFilename;
    #endif

    // Load the file
    GM_LayerHandle_t32 theLayerHandle = NULL;
    GM_Error_t32 theErr = GM_LoadGenericAsciiTextLayer
        (
        theFilename,
        &theLayerHandle,
        aLoadFlags,
        aFormatInfo,
        NULL
        );

    // Complete the load
    if ( NULL != theLayerHandle )
    {
        LoadLayerListImpl( &theLayerHandle, 1, aZoomToAllLayers );
    }
    else
    {
        // We failed to load the layer, show an error
        if ( GM_Error_NotRegistered == theErr )
        {
            ShowNotRegisteredMessage
                (
                _T("You are limited to loading a single layer without having a ")
                _T("registration file.")
                );
        }
        else if ( GM_Error_OperationCanceled != theErr )
        {
            TCHAR theErrStr[512];
            CString theFilename( aFilename );
            ::_stprintf
                ( 
                theErrStr, 
                _T("Error %d loading file %s"), 
                theErr, 
                (LPCTSTR)theFilename 
                );
            ShowMessageBox( theErrStr );
        }
    }

    // Free the filename if we allocated it
    #ifdef _UNICODE
    free( theFilename );
    theFilename = NULL;
    #endif

    return ( theErr );
} // end of function GM_MapWindow::LoadGenericAsciiTextLayer

//===================================================================
//
// METHOD:
//   GM_MapWindow::LoadLayer
//
// DESCRIPTION:
//   Loads the layers found in the given filename and adds them
//   to the list of layers to draw in this map window.
//
//   If <aLoadedLayersList> is not NULL, the handles for all of
//   the layers loaded will be added to that list.
//
//===================================================================
GM_Error_t32
GM_MapWindow::LoadLayer
    (
    LPCTSTR             aFilename,
    GM_LoadFlags_t32    aLoadFlags,
    LayerHandleList_t*  aLoadedLayersList,
    bool                aZoomToAllLayers
    )
{
    // If we're in a unicode build, get a multi-byte version of the filename
    #ifdef _UNICODE
    char* theFilename = ConvertUnicodeToMultibyte( aFilename );
    if ( NULL == theFilename )
    {
        return ( GM_Error_OutOfMemory );
    }
    #else
    const char* theFilename = aFilename;
    #endif

    // Load the file
    GM_LayerHandle_t32* theLayerHandleList = NULL;
    uint32 theNumLayersLoaded = 0;
    GM_Error_t32 theErr = LoadLayerImpl
        ( 
        theFilename, aLoadFlags, aLoadedLayersList, aZoomToAllLayers 
        );

    // Free the filename if we allocated it
    #ifdef _UNICODE
    free( theFilename );
    theFilename = NULL;
    #endif

    return ( theErr );
} // end of function GM_MapWindow::LoadLayer

//===================================================================
//
// METHOD:
//   GM_MapWindow::LoadLayer
//
// DESCRIPTION:
//   Prompts the user to select a file to load, then loads the 
//   layers found in the selected filename and adds them
//   to the list of layers to draw in this map window.
//
//   If <aLoadedLayersList> is not NULL, the handles for all of
//   the layers loaded will be added to that list.
//
//===================================================================
GM_Error_t32
GM_MapWindow::LoadLayer
    (
    GM_LoadFlags_t32    aLoadFlags,
    LayerHandleList_t*  aLoadedLayersList,
    bool                aZoomToAllLayers
    )
{
    // Prompt the user to select a file
    char theFilename[_MAX_PATH];
    GM_Error_t32 theErr = GM_SelectFile( theFilename, NULL, NULL, mDrawWnd );
    if ( GM_Error_None != theErr )
    {
        return ( theErr );
    }

    return ( LoadLayerImpl( theFilename, aLoadFlags, aLoadedLayersList, aZoomToAllLayers ) );
} // end of function GM_MapWindow::LoadLayer

//===================================================================
//
// METHOD:
//   GM_MapWindow::LoadOnlineLayer
//
// DESCRIPTION:
//   Loads a layer from an online data source, like WMS,
//   TerraServer-USA, or OpenStreetMap.org.
//
//===================================================================
GM_Error_t32
GM_MapWindow::LoadOnlineLayer
    (
    const char*             aThemeName,     // name of the theme to load ("drg", "doq", etc.) or NULL to ask user
    const GM_Rectangle_t*   aLatLonRect,    // lat/lon rect to import (NULL to ask user)
    bool                    aZoomToAll      // zoom to all layers when the load is complete
    )
{
    // Load the data. We provide NULL for both the theme name and rectangle
    // so the user will just be prompted for this information.
    GM_LayerHandle_t32 theLayerHandle = NULL;
    GM_Error_t32 theErr = GM_LoadOnlineLayer
        ( 
        aThemeName, aLatLonRect, &theLayerHandle, NULL
        );

    // Add the layer to the list or log the error
    if ( NULL != theLayerHandle )
    {
        // Add the layer to the layer list
        mLayerHandleList.push_back( theLayerHandle );

        // Reset the 3D view rectangle so it will get redrawn
        memset( &mLast3DViewRect, 0, sizeof mLast3DViewRect );

        // Zoom so that all layers are visible if requested, otherwise
        // just redraw.
        if ( aZoomToAll )
        {
            ZoomToAllLayers();
        }
        else
        {
            Invalidate();
        }
    }
    else
    {
        // We failed to load the layer, show an error
        if ( GM_Error_NotRegistered == theErr )
        {
            ShowNotRegisteredMessage
                (
                _T("You are limited to loading a single layer without having a ")
                _T("registration file.")
                );
        }
        else
        {
            TCHAR theErrStr[256];
            ::_stprintf( theErrStr, _T("Error %d loading online data"), theErr );
            ShowMessageBox( theErrStr );
        }
    }

    return ( theErr );
} // end of function GM_MapWindow::LoadOnlineLayer

//===================================================================
//
// METHOD:
//   GM_MapWindow::SetRasterDisplayOptions
//
// DESCRIPTION:
//   Sets the current raster display options for the given layer.
//   If you just want to change a single option, the best approach
//   is to first call GetRasterDisplayOptions to get the current
//   settings, then modify the setting that you want to change and
//   call SetRasterDisplayOptions.
//
//===================================================================
GM_Error_t32
GM_MapWindow::SetRasterDisplayOptions
    (
    GM_LayerHandle_t32          aLayer,
    GM_RasterDisplayOptions_t&  aDisplayOpts
    )
{
    aDisplayOpts.mSize = sizeof( GM_RasterDisplayOptions_t );
    GM_Error_t32 theErr = GM_SetRasterDisplayOptions( aLayer, &aDisplayOpts );
    if ( GM_Error_None == theErr )
    {
        // Reset the last 3D view rectangle so it will be redrawn
        memset( &mLast3DViewRect, 0, sizeof mLast3DViewRect );

        // Redraw
        Invalidate();
    }
    return ( theErr );
} // end of function GM_MapWindow::SetRasterDisplayOptions

//===================================================================
//
// METHOD:
//   GM_MapWindow::SetViewProjection
//
// DESCRIPTION:
//   Sets the projection to view/export data in.
//
//===================================================================
GM_Error_t32
GM_MapWindow::SetViewProjection
    (
    const GM_Projection_t*  aProj
    )
{
    GM_Error_t32 theErr = GM_SetProjection( aProj );
    if ( GM_Error_None == theErr )
    {
        // Reset the last 3D view rectangle so it will be redrawn
        memset( &mLast3DViewRect, 0, sizeof mLast3DViewRect );

        // Zoom to all layers. This will also recalculate our view rectangle
        // in the new projection and redraw.
        ZoomToAllLayers();
    }

    return ( theErr );
} // end of function GM_MapWindow::SetViewProjection

//===================================================================
//
// METHOD:
//   GM_MapWindow::Show3DView
//
// DESCRIPTION:
//   Displays the 3D view window with any loaded data shown.
//   Returns the error code returned by the GM_3DSetView function.
//
//===================================================================
GM_Error_t32
GM_MapWindow::Show3DView()
{
    // Set the callback function to use when the 3D view requests a new
    // data view
    GM_3DSetChangeViewCallback( Change3DViewCallback, this );

    // Show the 3D view
    GM_Error_t32 theErr = GM_3DSetView
        ( 
        NULL,       // draw all loaded layers,
        0,  
        GM_DrawFlags_EraseBackground,
        &mCurViewRect, 
        NULL, 
        0 
        );

    // We failed to load the layer, show an error. Otherwise, save the
    // new 3D view window.
    if ( GM_Error_NotRegistered == theErr )
    {
        ShowNotRegisteredMessage
            (
            _T("You cannot display the 3D view without having a ")
            _T("registration file.")
            );
    }
    else if ( GM_Error_None == theErr )
    {
        mLast3DViewRect = mCurViewRect;
    }

    return ( theErr );
} // end of function GM_MapWindow::Hide3DView

//===================================================================
//
// METHOD:
//   GM_MapWindow::UnloadAllLayers
//
// DESCRIPTION:
//   Unloads all layers that are currently open.
//
//===================================================================
void
GM_MapWindow::UnloadAllLayers
    (
    bool    aRedraw
    )
{
    // Close any layers that are still open
    for ( uint32 i = 0; i < mLayerHandleList.size(); i++ )
    {
        GM_CloseLayer( mLayerHandleList[i] );
    }
    mLayerHandleList.clear();

    // Redraw if necessary
    if ( aRedraw )
    {
        // Reset the last 3D view rectangle so it will be redrawn
        memset( &mLast3DViewRect, 0, sizeof mLast3DViewRect );

        // Zoom to all layers
        ZoomToAllLayers();
    }
} // end of function GM_MapWindow::UnloadAllLayers

//===================================================================
//
// METHOD:
//   GM_MapWindow::UnloadLayer
//
// DESCRIPTION:
//   Unloads a specific layer.
//
//===================================================================
GM_Error_t32
GM_MapWindow::UnloadLayer
    (
    GM_LayerHandle_t32  aLayer
    )
{
    // Make sure this layer was loaded with this map window
    LayerHandleList_t::iterator i = std::find
        (
        mLayerHandleList.begin(), mLayerHandleList.end(), aLayer
        );
    if ( i == mLayerHandleList.end() )
    {
        return ( GM_Error_InvalidLayer );
    }

    // Close the layer
    GM_Error_t32 theErr = GM_CloseLayer( aLayer );
    
    // Remove the layer from the layer list
    mLayerHandleList.erase( i );

    // Reset the last 3D view rectangle so it will be redrawn
    memset( &mLast3DViewRect, 0, sizeof mLast3DViewRect );

    return ( theErr );
} // end of function GM_MapWindow::UnloadLayer

//===================================================================
//
// METHOD:
//   GM_MapWindow::ZoomIn
//
// DESCRIPTION:
//   Zooms in by the given factor.
//
//===================================================================
void
GM_MapWindow::ZoomIn
    (
    double  aFactor
    ) 
{
    // Make sure the zoom factor makes sense
    if ( aFactor < 1.0 )
    {
        return;
    }

    // Get the width and height of the view rect
    double theWidth = mCurViewRect.mMaxX - mCurViewRect.mMinX;
    double theHeight = mCurViewRect.mMaxY - mCurViewRect.mMinY;

    // Get the center of the rectangle
    double theCenterX = ( mCurViewRect.mMaxX + mCurViewRect.mMinX ) / 2.0;
    double theCenterY = ( mCurViewRect.mMaxY + mCurViewRect.mMinY ) / 2.0;

    // Calculate the new width and height
    double theNewWidth = theWidth / aFactor;
    double theNewHeight = theHeight / aFactor;

    // Update the view rectangle size
    mCurViewRect.mMinX = theCenterX - ( theNewWidth / 2.0 );
    mCurViewRect.mMaxX = mCurViewRect.mMinX + theNewWidth;
    mCurViewRect.mMinY = theCenterY - ( theNewHeight / 2.0 );
    mCurViewRect.mMaxY = mCurViewRect.mMinY + theNewHeight;

    // Redraw
    Invalidate();
} // end of function GM_MapWindow::ZoomIn

//===================================================================
//
// METHOD:
//   GM_MapWindow::ZoomOut
//
// DESCRIPTION:
//   Zooms out by the given factor.
//
//===================================================================
void 
GM_MapWindow::ZoomOut
    (
    double  aFactor
    ) 
{
    // Make sure the zoom factor makes sense
    if ( aFactor < 1.0 )
    {
        return;
    }

    // Get the width and height of the view rect
    double theWidth = mCurViewRect.mMaxX - mCurViewRect.mMinX;
    double theHeight = mCurViewRect.mMaxY - mCurViewRect.mMinY;

    // Get the center of the rectangle
    double theCenterX = ( mCurViewRect.mMaxX + mCurViewRect.mMinX ) / 2.0;
    double theCenterY = ( mCurViewRect.mMaxY + mCurViewRect.mMinY ) / 2.0;

    // Calculate the new width and height
    double theNewWidth = theWidth * aFactor;
    double theNewHeight = theHeight * aFactor;

    // Update the view rectangle size
    mCurViewRect.mMinX = theCenterX - ( theNewWidth / 2.0 );
    mCurViewRect.mMaxX = mCurViewRect.mMinX + theNewWidth;
    mCurViewRect.mMinY = theCenterY - ( theNewHeight / 2.0 );
    mCurViewRect.mMaxY = mCurViewRect.mMinY + theNewHeight;

    // Redraw
    Invalidate();
} // end of function GM_MapWindow::ZoomOut

//===================================================================
//
// METHOD:
//   GM_MapWindow::ZoomToAllLayers
//
// DESCRIPTION:
//   Zooms the view to all loaded layers.
//
//===================================================================
void
GM_MapWindow::ZoomToAllLayers()
{
    // Add each bounding box to the global box
    memset( &mCurViewRect, 0, sizeof mCurViewRect );
    for ( uint32 i = 0; i < mLayerHandleList.size(); i++ )
    {
        GM_LayerHandle_t32 theLayer = mLayerHandleList[i];

        // Get info about the layer
        const GM_LayerInfo_t* theLayerInfo = GM_GetLayerInfo( theLayer );
        if ( NULL == theLayerInfo )
        {
            ShowMessageBox( _T("Error getting layer info") );
            continue;
        }

        // Add the global bounds to this one
        if ( 0 == i )
        {
            mCurViewRect = theLayerInfo->mGlobalRect;
        }
        else
        {
            mCurViewRect.mMinX = 
                MIN_VAL( mCurViewRect.mMinX, theLayerInfo->mGlobalRect.mMinX );
            mCurViewRect.mMinY = 
                MIN_VAL( mCurViewRect.mMinY, theLayerInfo->mGlobalRect.mMinY );
            mCurViewRect.mMaxX = 
                MAX_VAL( mCurViewRect.mMaxX, theLayerInfo->mGlobalRect.mMaxX );
            mCurViewRect.mMaxY = 
                MAX_VAL( mCurViewRect.mMaxY, theLayerInfo->mGlobalRect.mMaxY );
        }
    }

    // Expand the view rectangle a bit to get some padding around the data
    double theWidth = mCurViewRect.mMaxX - mCurViewRect.mMinX;
    double theHeight = mCurViewRect.mMaxY - mCurViewRect.mMinY;
    if ( 0.0 == theWidth )
    {
        theWidth = 1.0;
    }
    if ( 0.0 == theHeight )
    {
        theHeight = 1.0;
    }
    mCurViewRect.mMinX -= theWidth * 0.05;
    mCurViewRect.mMaxX += theWidth * 0.05;
    mCurViewRect.mMinY -= theHeight * 0.05;
    mCurViewRect.mMaxY += theHeight * 0.05;

    // Redraw
    Invalidate();
} // end of function GM_MapWindow::ZoomToAllLayers

/*--------------------------------------------------------------------
                 PUBLIC STATIC METHODS - UNICODE ONLY
--------------------------------------------------------------------*/

#ifdef _UNICODE

//===================================================================
//
// METHOD:
//   GM_MapWindow::ConvertMultibyteToUnicode
//
// DESCRIPTION:
//   Converts a multi-byte character string to a Unicode string.
//   The pointer returned is on the heap and must be de-allocated
//   with a call to free when the caller is done with it. If an
//   error occurs converting the string, NULL is returned.
//
//===================================================================
WCHAR*
GM_MapWindow::ConvertMultibyteToUnicode
    (
    const char* aStr
    )
{
    // Local variables
    int theConvertedCnt;
    int theCurCodePage;
    int theWideCharCnt;
    WCHAR* theUnicodeString = NULL;

    // Validate the input
    if ( NULL == aStr )
    {
        return ( NULL );
    }

    // TODO: Get the code page of the current language
    theCurCodePage = 1252;

    // Get a count of how many Unicode characters will be required for the
    // translated string
    theWideCharCnt = MultiByteToWideChar
        (
        theCurCodePage,
        0,
        aStr,
        -1,
        NULL,
        0
        );
    if ( theWideCharCnt <= 0 )
    {
        return ( NULL );
    }

    // Allocate a buffer for holding the wide characters
    theUnicodeString = (WCHAR*)malloc( theWideCharCnt * sizeof( WCHAR ) );
    if ( NULL == theUnicodeString )
    {
        return ( NULL );
    }

    // Perform the conversion
    theConvertedCnt = MultiByteToWideChar
        (
        theCurCodePage,
        0,
        aStr,
        -1,
        theUnicodeString,
        theWideCharCnt
        );
    if ( theConvertedCnt != theWideCharCnt )
    {
        free( theUnicodeString );
        return ( NULL );
    }

    // Return the unicode string
    return ( theUnicodeString );
} // end of function GM_MapWindow::ConvertMultibyteToUnicode

//===================================================================
//
// METHOD:
//   GM_MapWindow::ConvertUnicodeToMultibyte
//
// DESCRIPTION:
//   Converts a Unicode string to a multi-byte character string.
//   The pointer returned is on the heap and must be de-allocated
//   with a call to free when the caller is done with it. If an
//   error occurs converting the string, NULL is returned.
//
//===================================================================
char*
GM_MapWindow::ConvertUnicodeToMultibyte
    (
    const WCHAR*    aStr
    )
{
    // Local variables
    int theConvertedCnt;
    int theCurCodePage;
    int theMultByteCnt;
    char* theMultiByteString = NULL;

    // Validate the input
    if ( NULL == aStr )
    {
        return ( NULL );
    }

    // TODO: Get the code page of the current language
    theCurCodePage = 1252;

    // Get a count of how many multibyte characters will be required for the
    // translated string
    theMultByteCnt = WideCharToMultiByte
        (
        theCurCodePage,
        0,
        aStr,
        -1,
        NULL,
        0,
	    NULL,
	    NULL
        );
    if ( theMultByteCnt <= 0 )
    {
        return ( NULL );
    }

    // Allocate a buffer for holding the wide characters
    theMultiByteString = (char*)malloc( theMultByteCnt );
    if ( NULL == theMultiByteString )
    {
        return ( NULL );
    }

    // Perform the conversion
    theConvertedCnt = WideCharToMultiByte
        (
        theCurCodePage,
        0,
        aStr,
        -1,
        theMultiByteString,
        theMultByteCnt,
	    NULL,
	    NULL
        );
    if ( theConvertedCnt != theMultByteCnt )
    {
        free( theMultiByteString );
        return ( NULL );
    }

    // Return the multi-byte string
    return ( theMultiByteString );
} // end of function GM_MapWindow::ConvertUnicodeToMultibyte

#endif // defined _UNICODE

/*--------------------------------------------------------------------
                           PRIVATE METHODS
--------------------------------------------------------------------*/

//===================================================================
//
// METHOD:
//   GM_MapWindow::DrawToOffscreenBuffer
//
// DESCRIPTION:
//   Draws the loaded layers to the given device context using
//   the provided draw order. You must have called SetDrawWindowRect
//   at some point prior to calling this function. The drawing
//   is done to the background bitmap. When this is called the
//   background bitmap should already be allocated and selected
//   into the provided device context.
//
//===================================================================
GM_Error_t32
GM_MapWindow::DrawToOffscreenBuffer
    (
    HDC                 aDC,
    COLORREF            aBackgroundColor,
    DrawOrder_t         aDrawOrder,
    bool                aEraseBackground,
    GM_DrawFlags_t32    aDrawFlags
    )
{
    // Set the background color
    GM_SetBackgroundColor( aBackgroundColor );

    // Get the draw window size
    sint32 theDrawWndWidth = mDrawWndRect.right - mDrawWndRect.left;
    sint32 theDrawWndHeight = mDrawWndRect.bottom - mDrawWndRect.top;

    // If nothing is loaded, we're done
    GM_Error_t32 theErr = GM_Error_None;
    if ( mLayerHandleList.empty() )
    {
        if ( aEraseBackground )
        {
            EraseBackground( aDC, aBackgroundColor );
        }
        return ( theErr );
    }

    // Set the orientation
    GM_SetOrientation( mOrientation );

    // Draw the loaded layers based on the draw order
    if ( eDrawOrderRastersFirst == aDrawOrder )
    {
        // Allocate space for the list of vector layers and the list of raster
        // layers
        GM_LayerHandle_t32* theRasterLayers = 
            new GM_LayerHandle_t32[ mLayerHandleList.size() ];
        GM_LayerHandle_t32* theVectorLayers = 
            new GM_LayerHandle_t32[ mLayerHandleList.size() ];
        uint32 theVectorLayerCnt = 0;
        uint32 theRasterLayerCnt = 0;
        if ( ( NULL == theVectorLayers ) || ( NULL == theRasterLayers ) )
        {
            return ( GM_Error_OutOfMemory );
        }

        // Go through all loaded layers and draw the raster layers. Also
        // build a list of vector layers while we're doing that.
        for ( uint32 i = 0; ( i < mLayerHandleList.size() ) && ( GM_Error_None == theErr ); i++ )
        {
            GM_LayerHandle_t32 theLayerHandle = mLayerHandleList[i];

            // Get information about this layer
            const GM_LayerInfo_t* theLayerInfo = GM_GetLayerInfo( theLayerHandle );
            if ( NULL == theLayerInfo ) 
            {
                continue;
            }

            // If this is a raster layer, add it to the list of raster layers
            if ( theLayerInfo->mHasRasterData )
            {
                theRasterLayers[ theRasterLayerCnt ] = theLayerHandle;
                theRasterLayerCnt++;
            }
            else if ( theLayerInfo->mHasVectorData )
            {
                theVectorLayers[ theVectorLayerCnt ] = theLayerHandle;
                theVectorLayerCnt++;
            }
        }

        // Append the vector layers to the end of the raster layer list
        if ( theVectorLayerCnt > 0 )
        {
            memmove
                ( 
                &theRasterLayers[ theRasterLayerCnt ], 
                theVectorLayers, 
                theVectorLayerCnt * sizeof( GM_LayerHandle_t32 ) 
                );
        }

        // Draw the layers
        if ( ( ( theRasterLayerCnt + theVectorLayerCnt ) > 0 ) && 
             ( GM_Error_None == theErr ) )
        {
            // If we're supposed to erase the background, OR that flag in
            // with the draw flags
            if ( aEraseBackground )
            {
                aDrawFlags |= GM_DrawFlags_EraseBackground;
            }

            // Draw the layers
            theErr = GM_DrawLayerList
                ( 
                aDC,                    // device context to draw to
                theRasterLayers,        // list of layers to draw
                theRasterLayerCnt + theVectorLayerCnt,
                aDrawFlags,             // draw flags
                &mCurViewRect,          // rectangle in current projection to draw
                0,                      // draw to top-left of offscreen buffer
                0,                      // draw to top-left of offscreen buffer
                theDrawWndWidth,        // pixel width of draw window
                theDrawWndHeight        // pixel height of draw window
                );
        }
        else if ( aEraseBackground )
        {
            EraseBackground( aDC, aBackgroundColor );
        }

        // Free the vector and raster layer lists
        delete[] theRasterLayers;
        theRasterLayers = NULL;
        delete[] theVectorLayers;
        theVectorLayers = NULL;
    }
    else if ( eDrawOrderAsLoaded == aDrawOrder )
    {
        // We don't allow rotation in this draw order
        if ( 0.0 != mOrientation )
        {
            ShowMessageBox
                (
                _T("You cannot orient the map in any direction other than ")
                _T("north-up when using the eDrawOrderAsLoaded draw order.")
                );
            theErr = GM_Error_InvalidParam;
            return ( theErr );
        }

        // Erase the background if requested
        if ( aEraseBackground )
        {
            EraseBackground( aDC, aBackgroundColor );
        }

        // Draw all of the loaded layers one at a time
        for ( uint32 i = 0; ( i < mLayerHandleList.size() ) && ( GM_Error_None == theErr ); i++ )
        {
            GM_LayerHandle_t32 theLayerHandle = mLayerHandleList[i];
            theErr = GM_DrawLayerList
                ( 
                aDC, 
                &theLayerHandle,
                1,
                aDrawFlags,             // draw flags
                &mCurViewRect,
                0, 
                0, 
                theDrawWndWidth,
                theDrawWndHeight
                );   
            
            // Remove the erase background flag from the draw flags if provided
            aDrawFlags &= ~GM_DrawFlags_EraseBackground;
        }
    }
    else
    {
        theErr = GM_Error_InvalidParam;
    }

    // Draw grid lines if necessary
    if ( ( GM_Error_None == theErr ) && ( eNoGrid != mGridDisplay ) )
    {
        theErr = GM_DrawGridLinesEx
            (
            aDC,                        // device to draw to
            eLatLonGrid == mGridDisplay,
            0.0,                        // use the default grid spacing
            &mCurViewRect,              // view rectangle to draw
            0,
            0,
            theDrawWndWidth,
            theDrawWndHeight,
            NULL
            );
    }

    // If the 3D view window is open and we've changed views since the last
    // 3D update, update its view
    if ( GM_3DIsWindowOpen() &&
         ( 0 != memcmp( &mCurViewRect, &mLast3DViewRect, sizeof( GM_Rectangle_t ) ) ) )
    {
        GM_Error_t32 theErr = GM_3DSetView
            ( 
            NULL,       // draw all loaded layers,
            0,  
            GM_DrawFlags_EraseBackground,
            &mCurViewRect, 
            NULL, 
            0 
            );
        mLast3DViewRect = mCurViewRect;
    }

    // API Sample: Calculate the width and height of the last
    // draw world rectangle in meters:
    #ifdef GM_TEST_API
    if ( GM_Error_None == theErr )
    {
        double theWidthMeters, theHeightMeters;
        theErr = GM_CalcWorldRectSize( NULL, &theWidthMeters, &theHeightMeters );
    }
    #endif // defined GM_TEST_API

    // API Sample: Get the scale that we just drew at
    #ifdef GM_TEST_API
    if ( GM_Error_None == theErr )
    {
        double theScale = GM_GetLastDrawnScale( aDC );

        // Get scale rectangle. It should be close to the view rectangle
        // although it could be different since the DLL will auto-adjust
        // the aspect ratio of the view rectangle that we pass into the draw.
        GM_Rectangle_t theScaleRect;
        GM_Point_t theCenterPoint;
        theCenterPoint.mX = ( mCurViewRect.mMinX + mCurViewRect.mMaxX ) / 2.0;
        theCenterPoint.mY = ( mCurViewRect.mMinY + mCurViewRect.mMaxY ) / 2.0;
        theErr = GM_CalcScaleRectangle
            (
            &theCenterPoint, 
            theScale, 
            aDC, 
            theDrawWndWidth, 
            theDrawWndHeight, 
            &theScaleRect
            );
    }
    #endif // defined GM_TEST_API

    return ( theErr );
} // end of function GM_MapWindow::DrawToOffscreenBuffer

//===================================================================
//
// METHOD:
//   GM_MapWindow::EraseBackground
//
// DESCRIPTION:
//   Erases the background of the map with the given color.
//
//===================================================================
void
GM_MapWindow::EraseBackground
    (
    HDC                 aDC,
    COLORREF            aBackgroundColor
    )
{
    // Create the brush to use
    HBRUSH theBgBrush = ::CreateSolidBrush( aBackgroundColor );

    // Select the brush into the device context
    HGDIOBJ theOldBrush = ::SelectObject( aDC, theBgBrush );

    // Blank out the draw area
    ::PatBlt
        (
        aDC,
        0, 
        0, 
        mDrawWndRect.right - mDrawWndRect.left,
        mDrawWndRect.bottom - mDrawWndRect.top,
        PATCOPY
        );

    // Restore the old brush
    ::SelectObject( aDC, theOldBrush );

    // Destroy the background brush
    ::DeleteObject( theBgBrush );
} // end of function GM_MapWindow::EraseBackground

//===================================================================
//
// METHOD:
//   GM_MapWindow::Invalidate
//
// DESCRIPTION:
//   Invalidates the map area of the draw window.
//
//===================================================================
void
GM_MapWindow::Invalidate
    (
    bool    aInvalidateBgBitmap
    )
{
    // Destroy the background bitmap if necessary
    if ( aInvalidateBgBitmap && ( NULL != mDrawBgBitmap ) )
    {
        ::DeleteObject( mDrawBgBitmap );
        mDrawBgBitmap = NULL;
    }

    // Invalidate the draw window rectangle
    if ( NULL != mDrawWnd )
    {
        ::InvalidateRect( mDrawWnd, &mDrawWndRect, TRUE );
    }
} // end of function GM_MapWindow::Invalidate

//===================================================================
//
// METHOD:
//   GM_MapWindow::LoadLayerImpl
//
// DESCRIPTION:
//   Loads the layers found in the given filename and adds them
//   to the list of layers to draw in this map window.
//
//   If <aLoadedLayersList> is not NULL, the handles for all of
//   the layers loaded will be added to that list.
//
//===================================================================
GM_Error_t32
GM_MapWindow::LoadLayerImpl
    (
    const char*         aFilename,
    GM_LoadFlags_t32    aLoadFlags,
    LayerHandleList_t*  aLoadedLayersList,
    bool                aZoomToAllLayers
    )
{
    GM_Error_t32 theErr = GM_Error_None;

    // Load the file
    GM_LayerHandle_t32* theLayerHandleList = NULL;
    uint32 theNumLayersLoaded = 0;
    theErr = GM_LoadLayerList
        (
        aFilename, &theLayerHandleList, &theNumLayersLoaded, aLoadFlags
        );

    // If we succeeded, copy all of the layers
    if ( ( NULL != theLayerHandleList ) && ( theNumLayersLoaded > 0 ) )
    {
        // Copy all of the layers to the list
        if ( NULL != aLoadedLayersList )
        {
            for ( uint32 i = 0; i < theNumLayersLoaded; i++ )
            {
                aLoadedLayersList->push_back( theLayerHandleList[i] );
            }
        }

        // Call the remainder of the implementation
        LoadLayerListImpl
            ( 
            theLayerHandleList, theNumLayersLoaded, aZoomToAllLayers 
            );
    }
    else
    {
        // We failed to load the layer, show an error
        if ( GM_Error_NotRegistered == theErr )
        {
            ShowNotRegisteredMessage
                (
                _T("You are limited to loading a single layer without having a ")
                _T("registration file.")
                );
        }
        else if ( GM_Error_OperationCanceled != theErr )
        {
            TCHAR theErrStr[512];
            CString theFilename( aFilename );
            ::_stprintf
                ( 
                theErrStr, 
                _T("Error %d loading file %s"), 
                theErr, 
                (LPCTSTR)theFilename 
                );
            ShowMessageBox( theErrStr );
        }
    }

    return ( theErr );
} // end of function GM_MapWindow::LoadLayerImpl

//===================================================================
//
// METHOD:
//   GM_MapWindow::LoadLayerListImpl
//
// DESCRIPTION:
//   Does post-load setup for all of the layers in a layer list.
//
//===================================================================
void
GM_MapWindow::LoadLayerListImpl
    (
    GM_LayerHandle_t32* aLayerHandleList,
    uint32              aNumLayers,
    bool                aZoomToAllLayers
    )
{
    // Add each layer
    for ( uint32 i = 0; i < aNumLayers; i++ )
    {
        GM_LayerHandle_t32 theLayerHandle = aLayerHandleList[i];
        
        // Add the layer to the layer list
        mLayerHandleList.push_back( theLayerHandle );
        
        // API TEST: Test much of the API functionality.
        #ifdef GM_TEST_API
        TestAPI( theLayerHandle );
        #endif
    }
    
    // Reset the last 3D view rectangle so it will be redrawn
    memset( &mLast3DViewRect, 0, sizeof mLast3DViewRect );
    
    // Redraw
    if ( aZoomToAllLayers || ( aNumLayers == mLayerHandleList.size() ) )
    {
        ZoomToAllLayers();
    }
    else
    {
        Invalidate();
    }
} // end of function GM_MapWindow::LoadLayerListImpl
    
//===================================================================
//
// METHOD:
//   GM_MapWindow::ShowLayerDisplayOptionsDialog
//
// DESCRIPTION:
//   Displays a layer options dialog for the provided list of layers.
//
//===================================================================
void
GM_MapWindow::ShowLayerDisplayOptionsDialog
    (
    LayerHandleList_t&  aLayerList
    )
{
    // Show layer options
    GM_ShowLayerOptionsDialog( &aLayerList[0], aLayerList.size() );

    // Redraw
    Invalidate();
} // end of function GM_MapWindow::ShowLayerDisplayOptionsDialog

//===================================================================
//
// METHOD:
//   GM_MapWindow::ShowMessageBox
//
// DESCRIPTION:
//   Displays a message box with the given string.
//
//===================================================================
void
GM_MapWindow::ShowMessageBox
    (
    LPCTSTR aStr,
    UINT    aType
    ) const
{
    ::MessageBox( mDrawWnd, aStr, _T("Global Mapper"), aType );
} // end of function GM_MapWindow::ShowMessageBox

//===================================================================
//
// METHOD:
//   GM_MapWindow::ShowNotRegisteredMessage
//
// DESCRIPTION:
//   Displays a message when a feature requiring a registration
//   key is tried.
//
//===================================================================
void
GM_MapWindow::ShowNotRegisteredMessage
    (
    LPCTSTR aPrefixStr
    ) const
{
    // Display a message with instructions on how to get a registration file.
    // Use a shorter message on WinCE since the screen is much smaller.
    TCHAR theErrStr[2048];
    #ifndef _WIN32_WCE
    ::_stprintf
        (
        theErrStr,
        _T("%s\n\nYou can obtain a permanent registration file from the Developers page at ")
        _T("http://www.globalmapper.com. Email support@globalmapper.com if you have any ")
        _T("questions or if you would like a temporary registration file for testing out ")
        _T("the Global Mapper SDK.\n\n")
        _T("If you already have a gmdll_regkey.txt registration file, please ensure that ")
        _T("it is in the same folder as the GlobalMapperInterface.dll file."),
        aPrefixStr
        );
    #else   
    ::_stprintf
        (
        theErrStr,
        _T("%s\n\nYou can obtain a permanent registration file from the Developers page at ")
        _T("http://www.globalmapper.com. Email support@globalmapper.com if you have any ")
        _T("questions or if you would like a temporary registration file for testing out ")
        _T("the Global Mapper SDK.\n\n"),
        aPrefixStr
        );
    #endif
    ShowMessageBox( theErrStr );
} // end of function GM_MapWindow::ShowNotRegisteredMessage

/*--------------------------------------------------------------------
                        PRIVATE API TEST METHODS
--------------------------------------------------------------------*/

#ifdef GM_TEST_API

//===================================================================
//
// METHOD:
//   GM_MapWindow::TestAPI
//
// DESCRIPTION:
//   Demonstrates how to use many of the functions in the SDK API.
//
//===================================================================
void
GM_MapWindow::TestAPI
    (
    GM_LayerHandle_t32  aLayerHandle
    )
{
    GM_Error_t32 theErr;

    // Get information about the layer that we loaded
    const GM_LayerInfo_t* theLayerInfo = GM_GetLayerInfo( aLayerHandle );
    if ( NULL != theLayerInfo )
    {
        // Just to exercise the API, get the color and/or elevation of
        // the center of the overlay for raster/elevation layers. For
        // vector layers, log the number of features loaded. Some layers
        // can be both raster/elevation and vector, so don't make this
        // an else/if.
        if ( theLayerInfo->mMinElevation != theLayerInfo->mMaxElevation )
        {
            // This is an elevation layer, so get the elevation of the
            // center.
            float theElev = 0.0;
            theErr = GM_GetLocationElevation
                (
                aLayerHandle,
                ( theLayerInfo->mGlobalRect.mMinX + theLayerInfo->mGlobalRect.mMaxX ) / 2.0,
                ( theLayerInfo->mGlobalRect.mMinY + theLayerInfo->mGlobalRect.mMaxY ) / 2.0,
                &theElev
                );
            if ( ( GM_Error_None != theErr ) && ( GM_Error_NotRegistered != theErr ) )
            {
                TCHAR theErrStr[128];
                ::_stprintf( theErrStr, _T("Error %d getting center elevation"), theErr );
                ShowMessageBox( theErrStr );
            }
            
            // Get the elevation by pixel of the center
            if ( ( theLayerInfo->mPixelHeight > 0 ) && ( theLayerInfo->mPixelWidth > 0 ) )
            {
                theErr = GM_GetPixelElevation
                    (
                    aLayerHandle,
                    theLayerInfo->mPixelHeight / 2,
                    theLayerInfo->mPixelWidth / 2,
                    &theElev
                    );
                if ( ( GM_Error_None != theErr ) && ( GM_Error_NotRegistered != theErr ) )
                {
                    TCHAR theErrStr[128];
                    ::_stprintf( theErrStr, _T("Error %d getting center elevation by pixel"), theErr );
                    ShowMessageBox( theErrStr );
                }

                // Test getting a whole row of elevations. This is much faster than accessing the elevations
                // a pixel at a time if you need to get them all.
                float* theElevRow = new float[ theLayerInfo->mPixelWidth ];
                if ( NULL != theElevRow )
                {
                    theErr = GM_GetPixelElevationRow
                        (
                        aLayerHandle,
                        theLayerInfo->mPixelHeight / 2,
                        theElevRow,
                        -9999.9f
                        );
                    delete[] theElevRow;
                }
            }
            
            // API Sample: Get a path profile across the diagonal of the layer.
            float theElevList[1024];
            theErr = GM_GetPathProfile
                (
                aLayerHandle,
                theLayerInfo->mGlobalRect.mMinX,
                theLayerInfo->mGlobalRect.mMinY,
                theLayerInfo->mGlobalRect.mMaxX,
                theLayerInfo->mGlobalRect.mMaxY,
                theElevList,
                cnt_of_array( theElevList ),
                -9999.0
                );
            if ( ( GM_Error_None != theErr ) && ( GM_Error_NotRegistered != theErr ) )
            {
                TCHAR theErrStr[128];
                ::_stprintf( theErrStr, _T("Error %d getting path profile along diagonal"), theErr );
                ShowMessageBox( theErrStr );
            }

            // Calculate the width and height of the layer
            double theLayerWidth = 
                theLayerInfo->mGlobalRect.mMaxX - theLayerInfo->mGlobalRect.mMinX;
            double theLayerHeight = 
                theLayerInfo->mGlobalRect.mMaxY - theLayerInfo->mGlobalRect.mMinY;

            // API Sample: Get a path profile across the center half of the diagonal 
            // and perform a sample line-of-sight analysis.
            char thePathProfileLOSDetailsStr[1024];
            GM_PathProfileLOSParams_t thePathProfileLOSParams;
            ::memset( &thePathProfileLOSParams, 0, sizeof thePathProfileLOSParams );
            thePathProfileLOSParams.mSize = sizeof thePathProfileLOSParams;
            thePathProfileLOSParams.mFlags = GM_PathProfile_LOSValid;
            thePathProfileLOSParams.mStartX = theLayerInfo->mGlobalRect.mMinX + theLayerWidth / 4.0;
            thePathProfileLOSParams.mStartY = theLayerInfo->mGlobalRect.mMinY + theLayerHeight / 4.0;
            thePathProfileLOSParams.mEndX = thePathProfileLOSParams.mStartX + theLayerWidth / 2.0;
            thePathProfileLOSParams.mEndY = thePathProfileLOSParams.mStartY + theLayerHeight / 2.0;
            thePathProfileLOSParams.mElevList = theElevList;
            thePathProfileLOSParams.mListSize = cnt_of_array( theElevList );
            thePathProfileLOSParams.mDetailsStr = thePathProfileLOSDetailsStr;
            thePathProfileLOSParams.mDetailsStrMaxLen = sizeof( thePathProfileLOSDetailsStr );
            thePathProfileLOSParams.mAtmosphericCorr = 1.333333; // Standard atmospheric correct for radio waves
            thePathProfileLOSParams.mLOSFromHeight = 100.0; // Use line of sight from height 100.0 meters above ground
            thePathProfileLOSParams.mLOSToHeight = 150.0; // Use line of sight to height 150.0 meters above ground
            theErr = GM_GetPathProfileLOS( aLayerHandle, &thePathProfileLOSParams );
            if ( ( GM_Error_None != theErr ) && ( GM_Error_NotRegistered != theErr ) )
            {
                TCHAR theErrStr[128];
                ::_stprintf( theErrStr, _T("Error %d getting path profile/LOS along diagonal"), theErr );
                ShowMessageBox( theErrStr );
            }

            // API Sample: Apply an offset of 1.0 meters to all elevations
            GM_ElevationOverrides_t theElevOverrides;
            ::memset( &theElevOverrides, 0, sizeof theElevOverrides );
            theErr = GM_GetElevationOverrides( aLayerHandle, &theElevOverrides );
            if ( GM_Error_None == theErr )
            {
                theElevOverrides.mOffset = 1.0;
                theErr = GM_SetElevationOverrides( aLayerHandle, &theElevOverrides );
            }

            // API Sample: Create a custom shader and activate it

            // Define the custom shader
            GM_ElevColor_t theShaderElevList[] =
            {
                { 0.0,    RGB( 255, 0, 0 ) }, // Red at 0
                { 500.0,  RGB( 0, 255, 0 ) }, // Green at 500
                { 1000.0, RGB( 0, 0, 255 ) }, // Blue at 1000
            };
            GM_CustomShader_t theCustomShader;
            ::memset( &theCustomShader, 0, sizeof theCustomShader );
            theCustomShader.mShaderName = "API Test Shader";
            theCustomShader.mFlags = GM_CustomShader_StretchToRange;
            theCustomShader.mNumElevs = cnt_of_array( theShaderElevList );
            theCustomShader.mElevColorList = theShaderElevList;
            theErr = GM_AddCustomShader( &theCustomShader, 0 );
            if ( GM_Error_None == theErr )
            {
                // Now activate the shader
                GM_VerticalDisplayOptions_t theVertDispOpts =
                    *GM_GetVerticalDisplayOptions();
                theVertDispOpts.mShaderName = theCustomShader.mShaderName;
                theErr = GM_SetVerticalDisplayOptions( &theVertDispOpts );
            }

            // API Sample: Test calculation of area statistics

            // Initialize stats structure
            GM_AreaElevStats_t theAreaElevStats;
            ::memset( &theAreaElevStats, 0, sizeof theAreaElevStats );
            theAreaElevStats.mSize = sizeof theAreaElevStats;

            // Setup a list of points to calculate stats for
            GM_Point_t theAreaPts[3];
            theAreaPts[0].mX = theLayerInfo->mGlobalRect.mMinX + theLayerWidth / 2.0;
            theAreaPts[0].mY = theLayerInfo->mGlobalRect.mMinY + theLayerHeight / 2.0;
            theAreaPts[1].mX = theAreaPts[0].mX + theLayerWidth / 20.0;
            theAreaPts[1].mY = theAreaPts[0].mY + theLayerHeight / 20.0;
            theAreaPts[2].mX = theAreaPts[1].mX;
            theAreaPts[2].mY = theAreaPts[0].mY;

            // Calculate the stats. This is commented out now as it can be slow on some huge data sets,
            // just un-comment to test.
            /*
            theErr = GM_CalcAreaElevStats
                (
                &aLayerHandle,
                1,
                &theAreaElevStats,
                theAreaPts,
                cnt_of_array( theAreaPts ),
                0.0, // use default X spacing
                0.0, // use default Y spacing
                NULL
                );
            */

            // Uncomment to test contour generation (not supported on WinCE)
            //TestContourGeneration( aLayerHandle, theLayerInfo );
            
            // Uncomment to test view shed calculation (not supported on WinCE)
            //TestViewShedCalculation( theLayerInfo );

            // Uncomment to test volume calculation (not supported on WinCE)
            //TestVolumeCalculation( theLayerInfo );
        }
        
        if ( theLayerInfo->mPixelWidth > 0 )
        {
            // This is a raster layer, so get the color of the center.
            GM_Color_t theCenterColor;
            theErr = GM_GetLocationColor
                (
                aLayerHandle,
                ( theLayerInfo->mGlobalRect.mMinX + theLayerInfo->mGlobalRect.mMaxX ) / 2.0,
                ( theLayerInfo->mGlobalRect.mMinY + theLayerInfo->mGlobalRect.mMaxY ) / 2.0,
                &theCenterColor
                );
            if ( ( GM_Error_None != theErr ) && ( GM_Error_NotRegistered != theErr ) )
            {
                TCHAR theErrStr[128];
                ::_stprintf( theErrStr, _T("Error %d getting center color"), theErr );
                ShowMessageBox( theErrStr );
            }
            
            // Get the color by pixel of the center unless this is an elevation
            // layer
            if ( theLayerInfo->mMinElevation == theLayerInfo->mMaxElevation )
            {
                theErr = GM_GetPixelColor
                    (
                    aLayerHandle,
                    theLayerInfo->mPixelHeight / 2,
                    theLayerInfo->mPixelWidth / 2,
                    &theCenterColor
                    );
                if ( ( GM_Error_None != theErr ) && ( GM_Error_NotRegistered != theErr ) )
                {
                    TCHAR theErrStr[128];
                    ::_stprintf( theErrStr, _T("Error %d getting center color by pixel"), theErr );
                    ShowMessageBox( theErrStr );
                }
            }
        }
        
        // If the layer has vector data, log how much and also show a sample
        // of how to iterate through all of the features in the layer.
        if ( ( theLayerInfo->mNumAreas > 0 ) || 
             ( theLayerInfo->mNumLines > 0 ) ||
             ( theLayerInfo->mNumPoints > 0 ) )
        {
            // This layer has vector data, so log how much
            TCHAR theLogStr[128];
            ::_stprintf
                ( 
                theLogStr,
                _T("Loaded %d areas, %d lines, and %d points"),
                theLayerInfo->mNumAreas,
                theLayerInfo->mNumLines,
                theLayerInfo->mNumPoints
                );
            ShowMessageBox( theLogStr, MB_OK | MB_ICONINFORMATION );
            
            // API Sample: Get information about each area feature in the layer
            // Note that without a registration key, only the first
            // GM_MAX_NON_REGISTERED_VECTOR_IDX areas can be read successfully.
            if ( theLayerInfo->mNumAreas > 0 )
            {
                for ( uint32 i = 0; i < theLayerInfo->mNumAreas; i++ )
                {
                    // Get the area feature
                    GM_AreaFeature_t* theAreaFeature = GM_GetAreaFeature
                        (
                        aLayerHandle, i
                        );
                    if ( NULL == theAreaFeature )
                    {
                        break;
                    }
                    
                    // Get information about the area classification
                    GM_FeatureClassInfo_t theClassInfo;
                    GM_GetFeatureClassInfo
                        ( 
                        theAreaFeature->mFeatureInfo.mClass, 
                        GM_FeatureClass_Area, 
                        &theClassInfo 
                        );
                                        
                    // Test modification of area feature style. In this test change 
                    // all water areas to use a partially translucent (alpha-blended)
                    // blue color.
                    if ( ( theAreaFeature->mFeatureInfo.mClass >= AFC_MIN_WATER ) &&
                         ( theAreaFeature->mFeatureInfo.mClass <= AFC_MAX_WATER ) )
                    {
                        GM_AreaStyle_t theAreaStyle = theAreaFeature->mAreaStyle;

                        // Setup the color to be solid blue
                        theAreaStyle.mBrushStyle = GM_BRUSH_SOLID;
                        theAreaStyle.mBrushColor = RGB( 0, 0, 255 );
                        
                        // Add the alpha blend, which is in the upper 8 bits of
                        // the brush color
                        theAreaStyle.mBrushColor |= 128UL << 24;

                        // Set the area style
                        GM_SetAreaFeatureDrawStyle( aLayerHandle, i, &theAreaStyle );
                    }

                    // Calculate the enclosed area of the feature
                    double theEnclosedArea = 0.0;
                    GM_CalcEnclosedArea
                        (
                        theAreaFeature->mPointList, 
                        theAreaFeature->mNumPoints,
                        &theLayerInfo->mNativeProj,
                        &theEnclosedArea
                        );

                    // Free the area feature when you're done with it
                    GM_FreeAreaFeature( theAreaFeature );
                    theAreaFeature = NULL;
                }

                // API Sample: Create a buffer 10 meters around just the first area
                // in the layer and add it to this layer. Note this is ifdef'd out
                // since it can be slow.
                #if 0
                theErr = GM_CreateBufferArea
                    (
                    aLayerHandle, GM_FeatureClass_Area, 0, 10.0, aLayerHandle
                    );
                #endif
            }
            
            // API Sample: Get information about each line feature in the layer
            // Note that without a registration key, only the first
            // GM_MAX_NON_REGISTERED_VECTOR_IDX lines can be read successfully.
            if ( theLayerInfo->mNumLines > 0 )
            {
                for ( uint32 i = 0; i < theLayerInfo->mNumLines; i++ )
                {
                    // Get the line feature
                    GM_LineFeature_t* theLineFeature = GM_GetLineFeature
                        (
                        aLayerHandle, i
                        );
                    if ( NULL == theLineFeature )
                    {
                        continue;
                    }
                    
                    // Get information about the line classification
                    GM_FeatureClassInfo_t theClassInfo;
                    GM_GetFeatureClassInfo
                        ( 
                        theLineFeature->mFeatureInfo.mClass, 
                        GM_FeatureClass_Line, 
                        &theClassInfo 
                        );
                    
                    // Test modification of line feature style. In this example,
                    // bold the font of any major contour types and use a thicker
                    // line for those at a height above 2000. Made the labels for
                    // all contour lines be drawn on the contours lines instead
                    // of offset from them.
                    if ( ( theLineFeature->mFeatureInfo.mClass >= LFC_MIN_CONTOUR ) &&
                         ( theLineFeature->mFeatureInfo.mClass <= LFC_MAX_CONTOUR ) )
                    {
                        GM_LineStyle_t theLineStyle = theLineFeature->mLineStyle;

                        // Set the labels to draw on the line features
                        theLineStyle.mDrawLabelOnLine = TRUE;

                        // If it's a major contour (i.e. every 10th one), make it
                        // use a bold, italic font
                        if ( LFC_CONTOUR_MAJOR == theLineFeature->mFeatureInfo.mClass )
                        {
                            theLineStyle.mFont.mWeight = FW_BOLD;
                            theLineStyle.mFont.mItalicize = TRUE;
                        }

                        // Check for contours above 2000 and make them thicker
                        if ( ( NULL != theLineFeature->mFeatureInfo.mName ) &&
                             ( ::atoi( theLineFeature->mFeatureInfo.mName ) > 2000 ) )
                        {
                            theLineStyle.mPenWidth *= 2;
                        }

                        // Update the drawing style
                        GM_SetLineFeatureDrawStyle( aLayerHandle, i, &theLineStyle );                        
                    }

                    // Free the line feature when you're done with it
                    GM_FreeLineFeature( theLineFeature );
                    theLineFeature = NULL;
                }
            }
            
            // API Sample: Get information about each point feature in the layer.
            // Note that without a registration key, only the first
            // GM_MAX_NON_REGISTERED_VECTOR_IDX points can be read successfully.
            if ( theLayerInfo->mNumPoints > 0 )
            {
                for ( uint32 i = 0; i < theLayerInfo->mNumPoints; i++ )
                {
                    // Get the point feature
                    GM_PointFeature_t* thePointFeature = GM_GetPointFeature
                        (
                        aLayerHandle, i
                        );
                    if ( NULL == thePointFeature )
                    {
                        continue;
                    }
                    
                    // Get information about the point classification
                    GM_FeatureClassInfo_t theClassInfo;
                    GM_GetFeatureClassInfo
                        ( 
                        thePointFeature->mFeatureInfo.mClass, 
                        GM_FeatureClass_Point, 
                        &theClassInfo 
                        );
                    
                    // Test modification of point feature style. In this test
                    // change all spot elevation points to use a Dot symbol
                    // and a 14-pt underlined "Arial" font with red letters.
                    if ( PFC_SPOT_ELEVATION == thePointFeature->mFeatureInfo.mClass )
                    {
                        GM_PointStyle_t thePointStyle = thePointFeature->mPointStyle;

                        // Change the symbol
                        thePointStyle.mSymbolName = "Dot";

                        // Specify the font we want
                        thePointStyle.mFont.mFaceName = "Arial";
                        thePointStyle.mFont.mColor = RGB( 255, 0, 0 );
                        thePointStyle.mFont.mPointSize = 14;
                        thePointStyle.mFont.mUnderline = TRUE;

                        // Set the point style
                        GM_SetPointFeatureDrawStyle( aLayerHandle, i, &thePointStyle );
                    }

                    // Free the point feature when you're done with it
                    GM_FreePointFeature( thePointFeature );
                    thePointFeature = NULL;
                }
            }
        }
            
        // API Sample: Get the map catalog info. This will fail if it's not a 
        // map catalog which is ok.
        GM_MapCatalogInfo_t theCatalogInfo;
        ::memset( &theCatalogInfo, 0, sizeof theCatalogInfo );
        theCatalogInfo.mSize = sizeof theCatalogInfo;
        theErr = GM_MapCatalogGetInfo( aLayerHandle, &theCatalogInfo );
        for ( uint32 k = 0; k < theCatalogInfo.mNumMaps; k++ )
        {
            // Get information about this map in the catalog.
            const GM_LayerInfo_t* theMapInfo = 
                GM_MapCatalogGetMapInfo( aLayerHandle, k );
        }

        // API Sample: Project the center of the layer to lat/lon/WGS84
        double theCenterX = ( theLayerInfo->mGlobalRect.mMinX + theLayerInfo->mGlobalRect.mMaxX ) / 2.0;
        double theCenterY = ( theLayerInfo->mGlobalRect.mMinY + theLayerInfo->mGlobalRect.mMaxY ) / 2.0;
        double theCenterLat, theCenterLon;
        GM_Projection_t theGeoProj;
        ::memset( &theGeoProj, 0, sizeof theGeoProj );
        theGeoProj.mProjSys = GM_PRJ_GEO;
        theGeoProj.mDatum = GM_DATUM_WGS_84;
        theGeoProj.mUnit = GM_PRJ_UNIT_ARC_DEGREES;
        theErr = GM_ProjectPoint
            (
            theCenterX,
            theCenterY,
            &theCenterLon,
            &theCenterLat,
            NULL,               // project from current global coordinates
            &theGeoProj
            );
        if ( ( GM_Error_None != theErr ) && ( GM_Error_NotRegistered != theErr ) )
        {
            TCHAR theErrStr[128];
            ::_stprintf
                ( 
                theErrStr, 
                _T("Error %d projecting center coordinate to lat/lon"), 
                theErr 
                );
            ShowMessageBox( theErrStr );
        }
    }

    // API Sample: Change major grid lines to be colored in red and labeled with
    // a bold font.
    GM_LineStyle_t theGridLineStyle;
    ::memset( &theGridLineStyle, 0, sizeof theGridLineStyle );
    GM_GetLineFeatureClassDrawStyle( LFC_GRID_LINE_MAJOR, &theGridLineStyle );
    theGridLineStyle.mPenColor = RGB(255,0,0);
    theGridLineStyle.mFont.mWeight = FW_BOLD;
    GM_SetLineFeatureClassDrawStyle( LFC_GRID_LINE_MAJOR, &theGridLineStyle );
} // end of function GM_MapWindow::TestAPI

//===================================================================
//
// METHOD:
//   GM_MapWindow::TestContourGeneration
//
// DESCRIPTION:
//   Demonstrates how to use the API to generate contours from
//   a loaded elevation layer.
//
//===================================================================
void
GM_MapWindow::TestContourGeneration
    (
    GM_LayerHandle_t32      aLayerHandle,
    const GM_LayerInfo_t*   aLastLayerInfo
    )
{
    #ifndef _WIN32_WCE

    // API Sample: Generate contours over the entire layer

    // Setup the contour generation parameter structure
    GM_ContourParams_t theContourParams;
    ::memset( &theContourParams, 0, sizeof theContourParams );
    theContourParams.mSize = sizeof theContourParams;
    theContourParams.mDesc = "Generated Contours Test";
    theContourParams.mIntervalInFeet = FALSE;
    theContourParams.mGenerateAreas = FALSE;
    theContourParams.mGenerateSpotElevs = TRUE;
    theContourParams.mNumberOnlyLabels = FALSE;
    theContourParams.mShowProgress = TRUE;
    theContourParams.mXSpacing = 0.0; // Use max resolution of loaded data
    theContourParams.mYSpacing = 0.0; // Use max resolution of loaded data
    theContourParams.mSimpThreshold = 0.0; // Don't perform any simplification other than removing points along straight lines

    // Determine a good contour interval based on the range of elevation
    // values in the layer.
    float theElevRange = 
        aLastLayerInfo->mMaxElevation - aLastLayerInfo->mMinElevation;
    if ( theElevRange < 200.0 )
    {
        theContourParams.mContourInterval = 5;
    }
    else if ( theElevRange < 400.0 )
    {
        theContourParams.mContourInterval = 10;
    }
    else if ( theElevRange < 800.0 )
    {
        theContourParams.mContourInterval = 20;
    }
    else if ( theElevRange < 1000.0 )
    {
        theContourParams.mContourInterval = 25;
    }
    else if ( theElevRange < 2000.0 )
    {
        theContourParams.mContourInterval = 50;
    }
    else if ( theElevRange < 4000.0 )
    {
        theContourParams.mContourInterval = 100;
    }
    else
    {
        theContourParams.mContourInterval = 200;
    }

    // Generate the contours
    GM_LayerHandle_t32 theContourLayer = NULL;
    GM_Error_t32 theErr = GM_GenerateContours
        ( 
        aLayerHandle, &theContourParams, &theContourLayer 
        );
    if ( GM_Error_None == theErr )
    {
        // Add the contour layer to the display list
        mLayerHandleList.push_back( theContourLayer );
    }

    #endif // not defined _WIN32_WCE

} // end of function GM_MapWindow::TestContourGeneration

//===================================================================
//
// METHOD:
//   GM_MapWindow::TestViewShedCalculation
//
// DESCRIPTION:
//   Demonstrates how to use the API to perform a view shed 
//   calculation on a loaded elevation layer.
//
//===================================================================
void
GM_MapWindow::TestViewShedCalculation
    (
    const GM_LayerInfo_t*   aLastLayerInfo
    )
{
    #ifndef _WIN32_WCE

    // API Sample: Calculate the width and height in meters of the
    // overlay
    double theWidthMeters = 0.0;
    double theHeightMeters = 0.0;
    GM_Error_t32 theErr = GM_CalcWorldRectSize
        ( 
        &aLastLayerInfo->mGlobalRect, &theWidthMeters, &theHeightMeters 
        );
    
    // API Sample: Perform a small view shed analysis at the center
    // of the elevation layer

    // Setup the view shed parameter structure
    GM_ViewShedParams_t theViewShedParams;
    ::memset( &theViewShedParams, 0, sizeof theViewShedParams );
    theViewShedParams.mSize = sizeof theViewShedParams;
    theViewShedParams.mAtmosphericCorr = 1.0;
    theViewShedParams.mCenterPoint.mX = 
        ( aLastLayerInfo->mGlobalRect.mMinX + aLastLayerInfo->mGlobalRect.mMaxX ) / 2.0;
    theViewShedParams.mCenterPoint.mY = 
        ( aLastLayerInfo->mGlobalRect.mMinY + aLastLayerInfo->mGlobalRect.mMaxY ) / 2.0;
    theViewShedParams.mColor = RGB( 255, 0, 0 );
    theViewShedParams.mDesc = "API Sample View Shed";
    theViewShedParams.mRadius = ( theWidthMeters + theHeightMeters ) / 10.0;
    theViewShedParams.mReceiverAngle = NULL;
    theViewShedParams.mReceiverHeight = 2.0; // 2 meters above the ground
    theViewShedParams.mReceiverHeightAbs = FALSE;
    theViewShedParams.mShowHiddenPoints = FALSE;
    theViewShedParams.mShowProgress = TRUE;
    theViewShedParams.mTransmitterHeight = 100.0; // 100 meters above the ground
    theViewShedParams.mTransmitterHeightAbs = FALSE;
    theViewShedParams.mUseEarthCurvature = TRUE;
    theViewShedParams.mVectorElevsAbs = FALSE; // Vector heights are relative to ground level
    theViewShedParams.mXSpacing = 0.0; // Use max resolution of loaded data
    theViewShedParams.mYSpacing = 0.0; // Use max resolution of loaded data

    // API Sample: To demonstrate the use of vector data to enhance view 
    // shed analysis, create a custom layer with some vector features with 
    // elevation data to obstruct some of the view.

    GM_LayerHandle_t32 theCustomLayer = GM_CreateCustomVectorLayer
        (
        "Custom Vector Layer for View Shed Testing", &aLastLayerInfo->mNativeProj
        );
    if ( NULL != theCustomLayer )
    {
        // Setup a building northeast of the center point that is 20 meters high
        double theAreaWidth = 
            ( aLastLayerInfo->mGlobalRect.mMaxX - aLastLayerInfo->mGlobalRect.mMinX ) / 32.0;
        double theAreaHeight = 
            ( aLastLayerInfo->mGlobalRect.mMaxY - aLastLayerInfo->mGlobalRect.mMinY ) / 32.0;
        double theAreaCenterX = theViewShedParams.mCenterPoint.mX + theAreaWidth;
        double theAreaCenterY = theViewShedParams.mCenterPoint.mY + theAreaHeight;
        GM_Point_t theAreaPoints[4] =
        {
            { theAreaCenterX - theAreaWidth  / 2.0,
              theAreaCenterY - theAreaHeight / 2.0 }, // lower left
            { theAreaCenterX - theAreaWidth  / 2.0,
              theAreaCenterY + theAreaHeight / 2.0 }, // upper left
            { theAreaCenterX + theAreaWidth  / 2.0,
              theAreaCenterY + theAreaHeight / 2.0 }, // top right
            { theAreaCenterX + theAreaWidth  / 2.0,
              theAreaCenterY - theAreaHeight / 2.0 }  // bottom right
        };

        // Setup an attribute value for the elevation
        CString theAttrNameStr( "ELEVATION" );
        CString theAttrValStr( "20 m" );
        GM_AttrValue_t theAttrVal;
        theAttrVal.mName = theAttrNameStr.GetBuffer( theAttrNameStr.GetLength() );
        theAttrVal.mVal = theAttrValStr.GetBuffer( theAttrValStr.GetLength() );

        // Setup the area structure
        GM_AreaFeature_t theBuilding;
        CString theName( "Test Building" );
        CString theDesc( "Building" );
        ::memset( &theBuilding, 0, sizeof theBuilding );
        theBuilding.mFeatureInfo.mName = theName.GetBuffer( theName.GetLength() );
        theBuilding.mFeatureInfo.mDesc = theDesc.GetBuffer( theDesc.GetLength() );
        theBuilding.mFeatureInfo.mClass = AFC_UNKNOWN;
        theBuilding.mFeatureInfo.mAttrList = &theAttrVal;
        theBuilding.mFeatureInfo.mNumAttrs = 1;
        theBuilding.mPointList = theAreaPoints;
        theBuilding.mNumPoints = cnt_of_array( theAreaPoints );

        // Add the building polygon to the custom layer
        theErr = GM_AddAreaToVectorLayer( theCustomLayer, &theBuilding, TRUE );
        theAttrNameStr.ReleaseBuffer();
        theAttrValStr.ReleaseBuffer();
        theName.ReleaseBuffer();
        theDesc.ReleaseBuffer();

        // Setup a point tower to the south that is 150 ft high
        GM_PointFeature_t theTower;
        theName = "Test Tower";
        ::memset( &theTower, 0, sizeof theTower );
        theTower.mFeatureInfo.mName = theName.GetBuffer( theName.GetLength() );
        theTower.mFeatureInfo.mDesc = NULL;
        theTower.mFeatureInfo.mClass = PFC_TOWER;
        theAttrValStr = "150 ft";
        theAttrVal.mName = theAttrNameStr.GetBuffer( theAttrNameStr.GetLength() );
        theAttrVal.mVal = theAttrValStr.GetBuffer( theAttrValStr.GetLength() );
        theTower.mFeatureInfo.mAttrList = &theAttrVal;
        theTower.mFeatureInfo.mNumAttrs = 1;
        theTower.mPos.mX = theViewShedParams.mCenterPoint.mX;
        theTower.mPos.mY = theViewShedParams.mCenterPoint.mY - theAreaHeight * 2.0;

        // Add the tower to the custom layer
        theErr = GM_AddPointToVectorLayer( theCustomLayer, &theTower, TRUE );
        theAttrNameStr.ReleaseBuffer();
        theAttrValStr.ReleaseBuffer();
        theName.ReleaseBuffer();

        // Setup the view shed to use the custom vector layer that we created
        theViewShedParams.mVectorLayerList = &theCustomLayer;
        theViewShedParams.mVectorLayerListCnt = 1;
    }

    // Calculate the view shed
    GM_LayerHandle_t32 theViewShedLayer = NULL;
    theErr = GM_CalcViewShed( NULL, &theViewShedParams, &theViewShedLayer );
    if ( GM_Error_None == theErr )
    {
        const GM_LayerInfo_t* theViewShedLayerInfo = 
            GM_GetLayerInfo( theViewShedLayer );
        if ( NULL != theViewShedLayerInfo )
        {
            // See if a pixel in the view shed is visible
            GM_Color_t theViewShedColor;
            theErr = GM_GetPixelColor
                (
                theViewShedLayer,
                theViewShedLayerInfo->mPixelHeight / 3,
                theViewShedLayerInfo->mPixelWidth / 3,
                &theViewShedColor
                );
            if ( GM_Error_None == theErr )
            {
                // We got a color, so that means that the receiver
                // was visible at this location.
            }
            else
            {
                // We didn't get a color, so the receiver must
                // not have been visible at this location.
            }
        }
        
        // Add the view shed layer to the display list
        mLayerHandleList.push_back( theViewShedLayer );
    }

    // Close the custom layer
    if ( NULL != theCustomLayer )
    {
        GM_CloseLayer( theCustomLayer );
        theCustomLayer = NULL;
    }

    #endif // not defined _WIN32_WCE

} // end of function GM_MapWindow::TestViewShedCalculation

//===================================================================
//
// METHOD:
//   GM_MapWindow::TestVolumeCalculation
//
// DESCRIPTION:
//   Demonstrates how to use the API to perform cut-and-fill
//   volume calculations on a loaded elevation layer.
//
//===================================================================
void
GM_MapWindow::TestVolumeCalculation
    (
    const GM_LayerInfo_t*   aLastLayerInfo
    )
{
    #ifndef _WIN32_WCE
    
    // API Sample: Perform a small area volume analysis near the center
    // of the elevation layer

    // Setup an area northeast of the center point
    double theAreaWidth = 
        ( aLastLayerInfo->mGlobalRect.mMaxX - aLastLayerInfo->mGlobalRect.mMinX ) / 32.0;
    double theAreaHeight = 
        ( aLastLayerInfo->mGlobalRect.mMaxY - aLastLayerInfo->mGlobalRect.mMinY ) / 32.0;
    double theAreaCenterX = 
        ( aLastLayerInfo->mGlobalRect.mMinX + aLastLayerInfo->mGlobalRect.mMaxX ) / 2.0;
    double theAreaCenterY =
        ( aLastLayerInfo->mGlobalRect.mMinY + aLastLayerInfo->mGlobalRect.mMaxY ) / 2.0;
    GM_Point_t theAreaPoints[4] =
    {
        { theAreaCenterX - theAreaWidth  / 2.0,
          theAreaCenterY - theAreaHeight / 2.0 }, // lower left
        { theAreaCenterX - theAreaWidth  / 2.0,
          theAreaCenterY + theAreaHeight / 2.0 }, // upper left
        { theAreaCenterX + theAreaWidth  / 2.0,
          theAreaCenterY + theAreaHeight / 2.0 }, // top right
        { theAreaCenterX + theAreaWidth  / 2.0,
          theAreaCenterY - theAreaHeight / 2.0 }  // bottom right
    };

    // Setup the volume parameter structure
    GM_AreaVolumeParams_t theAreaVolumeParams;
    ::memset( &theAreaVolumeParams, 0, sizeof theAreaVolumeParams );
    theAreaVolumeParams.mSize = sizeof theAreaVolumeParams;
    theAreaVolumeParams.mPoints = theAreaPoints;
    theAreaVolumeParams.mNumPoints = cnt_of_array( theAreaPoints );
    theAreaVolumeParams.mCutHeight = 
        (float)( ( aLastLayerInfo->mMinElevation + aLastLayerInfo->mMaxElevation ) / 2.0 );
    theAreaVolumeParams.mXSpacing = 0.0; // Use max resolution of loaded data
    theAreaVolumeParams.mYSpacing = 0.0; // Use max resolution of loaded data

    // Calculate the volume of the area
    double theAreaCutVolume = 0.0;
    double theAreaFillVolume = 0.0;
    GM_Error_t32 theErr = GM_CalcVolumeOfArea
        (
        NULL,                   // Use the best layer at each sample location
        &theAreaVolumeParams,
        &theAreaCutVolume,
        &theAreaFillVolume
        );
    
    // API Sample: Perform a volume analysis along a short line near the center
    // of the elevation layer

    // Setup a line going across the diagonal of the area that we created
    GM_Point_t theLinePoints[2] =
    {
        { theAreaCenterX - theAreaWidth  / 2.0,
          theAreaCenterY - theAreaHeight / 2.0 }, // lower left
        { theAreaCenterX + theAreaWidth  / 2.0,
          theAreaCenterY + theAreaHeight / 2.0 }, // top right
    };

    // Setup the elevation list. We'll make the start point 1 meter above ground
    // and the stop point 2 meters below ground
    float theCutHeights[2] = { 1.0, -2.0 };

    // Setup the volume parameter structure
    GM_LineVolumeParams_t theLineVolumeParams;
    ::memset( &theLineVolumeParams, 0, sizeof theLineVolumeParams );
    theLineVolumeParams.mSize = sizeof theLineVolumeParams;
    theLineVolumeParams.mPoints = theLinePoints;
    theLineVolumeParams.mCutHeights = theCutHeights;
    theLineVolumeParams.mNumPoints = cnt_of_array( theLinePoints );
    theLineVolumeParams.mCutHeightsAbsolute = FALSE; // cut heights are relative to ground level
    theLineVolumeParams.mCorridorWidth = 2.0; // the diameter to measure around the line is 2.0 m
    theLineVolumeParams.mXSpacing = 0.0; // Use max resolution of loaded data
    theLineVolumeParams.mYSpacing = 0.0; // Use max resolution of loaded data

    // Calculate the volume along the line
    double theLineCutVolume = 0.0;
    double theLineFillVolume = 0.0;
    theErr = GM_CalcVolumeAlongLine
        (
        NULL,                   // Use the best layer at each sample location
        &theLineVolumeParams,
        &theLineCutVolume,
        &theLineFillVolume
        );

    #endif // not defined _WIN32_WCE

} // end of function GM_MapWindow::TestVolumeCalculation

#endif // defined GM_TEST_API

// end of file GM_MapWindow.cpp