// GM_MapWindow.h - interface for GM_MapWindow class
//
// This class provides an easy to use class wrapper for handling
// the loading and display of maps loaded by the Global Mapper
// SDK to a user-provided device context.

#ifndef _GM_MAPWINDOW_H_
#define _GM_MAPWINDOW_H_

#include "GlobalMapperInterface.h"

#include <vector>

class GM_MapWindow
{
//-------------------------------------------------------------------
public: // TYPES
//-------------------------------------------------------------------

    // Enumeration of the drawing order options
    enum DrawOrder_t
    {
        eDrawOrderRastersFirst, // Draw in order layers were loaded, with the
                                // exception that all raster and elevation
                                // layers will be drawn before any vector
                                // data is displayed. Also, all vector layers
                                // will be drawn simultaneously to allow for
                                // proper area, line, and point layering.
        eDrawOrderAsLoaded,     // Draw in order layers were loaded.

        eNumDrawOrders,

        eDrawOrderDefault = eDrawOrderRastersFirst
    };

    // Enumeration for grid display types
    enum GridDisplay_t
    {
        eLatLonGrid,            // Draw a latitude/longitude grid
        eViewProjGrid,          // Draw a grid in the current view projection
        eNoGrid                 // Do not display a grid
    };

    // Type for holding a list of layer handles
    typedef std::vector< GM_LayerHandle_t32 > LayerHandleList_t;

    // Type for holding a list of search results
    typedef std::vector< GM_FoundFeature_t > FoundFeatureList_t;

//-------------------------------------------------------------------
public: // FUNCTIONS
//-------------------------------------------------------------------

    //---------------------------------------------------------------
    // Constructor
    GM_MapWindow();

    //---------------------------------------------------------------
    // Destructor
    virtual ~GM_MapWindow();

    //---------------------------------------------------------------
    // Adds a layer to the map window that was externally created
    void
    AddLayer
    (
        GM_LayerHandle_t32  aLayer
    )
        {
        LoadLayerListImpl( &aLayer, 1, false );
        }

    //---------------------------------------------------------------
    // Recenters the map on the given location in ground coordinates.
    void
    CenterOnGroundPoint
    (
        double  aX,
        double  aY
    );

    //---------------------------------------------------------------
    // Recenters the map on the given location in pixel coordinates.
    GM_Error_t32
    CenterOnPixel
    (
        sint32  aPixelX,
        sint32  aPixelY
    );

    //---------------------------------------------------------------
    // Convert a ground coordinate to a pixel coordinate
    GM_Error_t32
    ConvertCoordGroundToPixel
    (
        double  aGroundX,   // in: ground X coordinate
        double  aGroundY,   // in: ground Y coordinate
        sint32& aPixelX,    // out: x pixel coordinate in draw window
        sint32& aPixelY     // out: y pixel coordiante in draw window
    ) const;

    //---------------------------------------------------------------
    // Convert a lat/lon value to a pixel coordinate
    GM_Error_t32
    ConvertCoordLatLonToPixel
    (
        double  aLat,       // in: latitude in degrees in current view datum
        double  aLon,       // in: longitude in degrees in current view datum
        sint32& aPixelX,    // out: x pixel coordinate in draw window
        sint32& aPixelY     // out: y pixel coordiante in draw window
    ) const;

    //---------------------------------------------------------------
    // Convert a pixel coordinate to a ground coordinate
    GM_Error_t32
    ConvertCoordPixelToGround
    (
        sint32  aPixelX,    // in: x pixel coordinate in draw window
        sint32  aPixelY,    // in: y pixel coordiante in draw window
        double& aGroundX,   // out: ground X coordinate
        double& aGroundY   // out: ground Y coordinate
    ) const;

    //---------------------------------------------------------------
    // Convert a pixel coordinate to a lat/lon value
    GM_Error_t32
    ConvertCoordPixelToLatLon
    (
        sint32  aPixelX,    // in: x pixel coordinate in draw window
        sint32  aPixelY,    // in: y pixel coordiante in draw window
        double& aLat,       // out: latitude in degrees in current view datum
        double& aLon        // out: longitude in degrees in current view datum
    ) const;

    //---------------------------------------------------------------
    // Draw the loaded layers to the device context
    GM_Error_t32
    Draw
    (
        HDC                 aDC,
        COLORREF            aBackgroundColor = RGB(255,255,206),
        DrawOrder_t         aDrawOrder = eDrawOrderDefault,
        bool                aEraseBackground = true,
        GM_DrawFlags_t32    aDrawFlags = 0
    );

    //---------------------------------------------------------------
    // Finds the nearest loaded (and drawn) vector feature(s) to the
    // given pixel location.
    FoundFeatureList_t
    FindNearestFeatures
    (
        const POINT&        aSearchPt,              // pixel location to search at
        uint32              aSearchRadius,          // search radius in screen pixels
        uint32              aMaxFeatures = 1,       // max number of features to find
        GM_FindFlags_t32    aFindFlags = GM_FindFlags_FindAll
    );

    //---------------------------------------------------------------
    // Returns the draw rectangle in pixel coordinates
    const RECT&
    GetDrawWndRect() const
        {
        return ( mDrawWndRect );
        }

    //---------------------------------------------------------------
    // Returns the view rectangle in the current view projection.
    const GM_Rectangle_t&
    GetCurrentViewRect() const
        {
        return ( mCurViewRect );
        }

    //---------------------------------------------------------------
    // Returns the list of currently loaded layers
    const LayerHandleList_t&
    GetLayerList() const
        {
        return ( mLayerHandleList );
        }

    //---------------------------------------------------------------
    // Retrieve the display options for a given raster/elevation layer
    GM_Error_t32
    GetRasterDisplayOptions
    (
        GM_LayerHandle_t32          aLayer,
        GM_RasterDisplayOptions_t&  aDisplayOpts
    ) const;

    //---------------------------------------------------------------
    // Hides the 3D view window if one is shown
    void
    Hide3DView();

    //---------------------------------------------------------------
    // Mark the map area as invalid so it will redraw
    void
    Invalidate
    (
        bool    aInvalidateBgBitmap = true // should we invalidate the background bitmap for a full redraw?
    );

    //---------------------------------------------------------------
    // Loads an ASCII text file layer
    GM_Error_t32
    LoadGenericAsciiTextLayer
    (
        LPCTSTR                     aFilename,
        const GM_AsciiFormatInfo_t* aFormatInfo = NULL,
        GM_LoadFlags_t32            aLoadFlags = GM_LoadFlags_HideWarnings,
        GM_LayerHandle_t32*         aLoadedLayer = NULL,
        bool                        aZoomToAllLayers = true
    );

    //---------------------------------------------------------------
    // Loads all layers from the given file
    GM_Error_t32
    LoadLayer
    (
        LPCTSTR             aFilename,
        GM_LoadFlags_t32    aLoadFlags = GM_LoadFlags_HideWarnings,
        LayerHandleList_t*  aLoadedLayersList = NULL,
        bool                aZoomToAllLayers = true
    );

    //---------------------------------------------------------------
    // Prompts the user to select a file to load and then loads all
    // of the layers from that file.
    GM_Error_t32
    LoadLayer
    (
        GM_LoadFlags_t32    aLoadFlags = GM_LoadFlags_HideWarnings,
        LayerHandleList_t*  aLoadedLayersList = NULL,
        bool                aZoomToAllLayers = true
    );

    //---------------------------------------------------------------
    GM_Error_t32
    LoadOnlineLayer
    (
        const char*             aThemeName = NULL, // name of the theme to load ("drg", "doq", etc.) or NULL to ask user
        const GM_Rectangle_t*   aLatLonRect = NULL,// lat/lon rect to import (NULL to ask user)
        bool                    aZoomToAll = true  // zoom to all layers when the load is complete
    );

    //---------------------------------------------------------------
    // Loads data from the TerraServer
    GM_Error_t32
    LoadTerraServerLayer
    (
        const char*             aThemeName = NULL, // name of the theme to load ("drg", "doq", etc.) or NULL to ask user
        const GM_Rectangle_t*   aLatLonRect = NULL,// lat/lon rect to import (NULL to ask user)
        bool                    aZoomToAll = true  // zoom to all layers when the load is complete
    )
        {
        return ( LoadOnlineLayer( aThemeName, aLatLonRect, aZoomToAll ) );
        }

    //---------------------------------------------------------------
    // Sets the view rectangle to display in the current view projection.
    void
    SetCurrentViewRect
    (
        const GM_Rectangle_t&   aViewRect
    )
        {
        mCurViewRect = aViewRect;
        if ( mCurViewRect.mMinX == mCurViewRect.mMaxX )
            {
            mCurViewRect.mMinX -= 1.0;
            mCurViewRect.mMaxX += 1.0;
            }
        if ( mCurViewRect.mMinY == mCurViewRect.mMaxY )
            {
            mCurViewRect.mMinY -= 1.0;
            mCurViewRect.mMaxY += 1.0;
            }
        Invalidate();
        }

    //---------------------------------------------------------------
    // Sets the window to which we will be drawing. This is used
    // to invalidate it when it needs redrawn.
    void
    SetDrawWindow
    (
        HWND    aWndHandle
    )
        {
        mDrawWnd = aWndHandle;
        }

    //---------------------------------------------------------------
    // Set draw window rectangle in client coordinates of device
    // context that will be drawn to.
    void
    SetDrawWindowRect
    (
        const RECT& aWndRect
    )
        {
        mDrawWndRect = aWndRect;
        }

    //---------------------------------------------------------------
    void
    SetGridDisplayType
    (
        GridDisplay_t   aGridDisplay
    )
        {
        if ( aGridDisplay != mGridDisplay )
            {
            mGridDisplay = aGridDisplay;
            Invalidate();
            }
        }

    //---------------------------------------------------------------
    // Sets the display options to use for the given layer.
    GM_Error_t32
    SetRasterDisplayOptions
    (
        GM_LayerHandle_t32          aLayer,
        GM_RasterDisplayOptions_t&  aDisplayOpts
    );
        
    //---------------------------------------------------------------
    // Sets the cartographic angle to which the top of the map should
    // point. This means that a north-up map (the default) is 0, 
    // east-up is 90, south-up is 180, and west-up is 270.
    void
    SetOrientationAngle
    (
        double  aOrientation
    )
        {
        if ( mOrientation != aOrientation )
            {
            mOrientation = aOrientation;
            Invalidate();
            }
        }

    //---------------------------------------------------------------
    // Set the projection to view data in.
    GM_Error_t32
    SetViewProjection
    (
        const GM_Projection_t*  aProj
    );

    //---------------------------------------------------------------
    // Displays a 3D view on the loaded data (if any)
    GM_Error_t32
    Show3DView();

    //---------------------------------------------------------------
    void
    ShowLayerDisplayOptionsDialog
    (
        LayerHandleList_t&  aLayerList
    );

    //---------------------------------------------------------------
    // Displays a message when an operation cannot be performed
    // because a valid registration license could not be found.
    void
    ShowNotRegisteredMessage
    (
        LPCTSTR aPrefixStr
    ) const;

    //---------------------------------------------------------------
    // Unloads all previously loaded layers.
    void
    UnloadAllLayers
    (
        bool    aRedraw = true
    );

    //---------------------------------------------------------------
    // Unloads a specific layer.
    GM_Error_t32
    UnloadLayer
    (
        GM_LayerHandle_t32  aLayer
    );

    //---------------------------------------------------------------
    void
    ZoomIn
    (
        double  aFactor = 2.0
    );

    //---------------------------------------------------------------
    void
    ZoomOut
    (
        double  aFactor = 2.0
    );

    //---------------------------------------------------------------
    // Zoom the view such that all loaded layers are visible.
    void
    ZoomToAllLayers();

//-------------------------------------------------------------------
public: // STATIC FUNCTIONS - UNICODE ONLY
//-------------------------------------------------------------------

    #ifdef _UNICODE

    //---------------------------------------------------------------
    static WCHAR*
    ConvertMultibyteToUnicode
    (
        const char* aStr
    );

    //---------------------------------------------------------------
    static char*
    ConvertUnicodeToMultibyte
    (
        const WCHAR*    aStr
    );

    #endif // defined _UNICODE

//-------------------------------------------------------------------
private: // FUNCTIONS
//-------------------------------------------------------------------

    //---------------------------------------------------------------
    // Draw the loaded layers to the offscreen bitmap
    GM_Error_t32
    DrawToOffscreenBuffer
    (
        HDC                 aDC,
        COLORREF            aBackgroundColor,
        DrawOrder_t         aDrawOrder,
        bool                aEraseBackground,
        GM_DrawFlags_t32    aDrawFlags
    );
    
    //---------------------------------------------------------------
    void
    EraseBackground
    (
        HDC                 aDC,
        COLORREF            aBackgroundColor
    );

    //---------------------------------------------------------------
    // Loads all layers from the given file
    GM_Error_t32
    LoadLayerImpl
    (
        const char*         aFilename,
        GM_LoadFlags_t32    aLoadFlags = 0,
        LayerHandleList_t*  aLoadedLayersList = NULL,
        bool                aZoomToAllLayers = true
    );

    //---------------------------------------------------------------
    // Loads all layers from the given file
    void
    LoadLayerListImpl
    (
        GM_LayerHandle_t32* aLayerHandleList,
        uint32              aNumLayers,
        bool                aZoomToAllLayers
    );

    //---------------------------------------------------------------
    void
    ShowMessageBox
    (
        LPCTSTR aStr,
        UINT    aType = MB_OK | MB_ICONWARNING
    ) const;

//-------------------------------------------------------------------
private: // API TEST FUNCTIONS
//-------------------------------------------------------------------

    #ifdef GM_TEST_API

    //---------------------------------------------------------------
    void
    TestAPI
    (
        GM_LayerHandle_t32  aLayerHandle
    );

    //---------------------------------------------------------------
    void
    TestContourGeneration
    (
        GM_LayerHandle_t32      aLayerHandle,
        const GM_LayerInfo_t*   aLastLayerInfo
    );

    //---------------------------------------------------------------
    void
    TestViewShedCalculation
    (
        const GM_LayerInfo_t*   aLastLayerInfo
    );

    //---------------------------------------------------------------
    void
    TestVolumeCalculation
    (
        const GM_LayerInfo_t*   aLastLayerInfo
    );

    #endif // defined GM_TEST_API

//-------------------------------------------------------------------
private: // DATA
//-------------------------------------------------------------------

    GM_Rectangle_t      mCurViewRect;       // current view rectangle in global coordinates
    HBITMAP             mDrawBgBitmap;      // background bitmap for double-buffered drawing
    HWND                mDrawWnd;
    RECT                mDrawWndRect;       // map draw window rectangle in client coordinates
    GridDisplay_t       mGridDisplay;       // the type of grid to display
    GM_Rectangle_t      mLast3DViewRect;    // last 3D view rectangle
    LayerHandleList_t   mLayerHandleList;   // list of currently loaded layers
    double              mOrientation;       // orientation to display map

}; // end of class GM_MapWindow

#endif // end of file GM_MapWindow.h